# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# _number.py 
#  written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  August 2003/November 2022 (python3)
# 
"""smart and reliable number dictation

the number part of the grammar was initially provided by Joel Gould in
his grammar "winspch.py it is changed a bit, but essentially all sorts of numbers can be
dictated with his grammar.

For real use in other grammars, copy the things from his grammar into another grammar. See
for example "_lines.py" and "firefox browsing.py".
BUT: first try if the logic of _number simple.py is enough. If not, study this grammar.

BTW when dictating numbers in excel (doing my bookkeeping), I use this grammar  (or the "number simple") all the time!

QH september 2013: rewriting of the functions, ruling out optional command words. The optional word "and" has been removed
            (now say "three hundred twenty" instead of "three hundred and twenty")

QH211203: English numbers require more work: thirty three can be recognised as "33" or
as "30", "3".

QH050104: standardised things, and put functions in natlinkutilsbj, so that
other grammars can invoke the number grammar more easily.
"""
#pylint:disable=C0209, R0904, R0912, R0915
import natlink
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doKeystroke as keystroke
from dtactions.unimacro.unimacroactions import getMetaAction
from dtactions.unimacro import unimacroutils

import unimacro.natlinkutilsbj as natbj

from unimacro import iban  # special module for banknumber (European mainly I think)

# Note: lists number1to99 and number1to9 and n1-9 and n20-90 etc. are taken from function getNumberList in natlinkutilsbj

class NumberException(Exception):
    pass


ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):

    language = unimacroutils.getLanguage()
    #Step 1, choose one of next three grammar rules:
    # the <integer> rule comes from these grammar rules
    try:
        number_rules = natbj.numberGrammarTill999[language]
    except KeyError:
        print('take number grammar from "enx"')
        number_rules = natbj.numberGrammarTill999['enx']
    
    #number_rules = natbj.numberGrammarTill999999[language] # including thousands
    #number_rules = natbj.numberGrammar[language] #  including millions
    name = "number extended"
    # do not use "To" instead of "Through", because "To" sounds the same as "2"
    gramSpec = ["""
<testnumber1> exported = [<spacing>] (number <number>)+;
<testnumber2> exported = number <integer> Through <integer>;
<pagenumber> exported = page <integer> [Through <integer>];
<filenamelastpage> exported = filename last page;
<pair> exported = Pair <number> And <number>;
<number> = <integer> | Minus <integer> | <float> | Minus <float>;
<spacingnumber> exported = number (<spacing>|<number>)+;
<spacing> = {spacing};
"""]
    gramSpec.append("<banknumber> exported = IBAN {n00-99} {bankheader} <integer>;")
    gramSpec.append(number_rules)
    try:
        integer999 = natbj.numberGrammarTill999[language]
    except KeyError:
        print('take number grammar from "enx"')
        integer999 = natbj.numberGrammarTill999['enx']

#     ##  345567;345567
#     ## these are special rules for making numbers of specific length.
#     ## when they appear to be useful, I will try to make them available in other languages
#     ## and other cases too:
#         
#     amsterdamZuidRule = """<numberaztotal> exported =  <numberazone> | <numberaztwo> | <numberazone><numberaztwo>;
#     <numberazone>  = zesentachtig <numberthreedigits><numberthreedigits>;
#     <numberaztwo> = point (<__thousandslimited>  | <numbertwodigits><numbertwodigits>);
# <numbertwodigits> = {n0-9}{n0-9} | {n10-99};
# <numberthreedigits> = {n0-9}{n0-9}{n0-9} | <__hundredslimited> | {n0-9} <numbertwodigits>;
# <__hundredslimited> = hundred | <__2to9before> hundred | hundred <__1to99after> ;
# <__thousandslimited> = thousand | <__2to9before> thousand | thousand <__1to99after> |
#                         <__2to9before> thousand <__1to9after>;
# 
# <numberfourdigits> = <__thousandslimited>  | <numbertwodigits> <numbertwodigits> ;
# <numbersixdigits> = {n0-9}+ | {n0-9} <numbertwodigits>  {n0-9} <numbertwodigits> ;
# <__thousands> = duizend | (<__1to99>|<__hundreds>) duizend | duizend (<__1to99>|<__hundreds>) |
#                    (<__1to99>|<__hundreds>) duizend  (<__1to99>|<__hundreds>);
# <__2to9before> = {n2-9};
# <__1to99after> = {n1-99};
# <__1to9after> = {n1-9};
# 
# """

##8600 8650 86 2008
    def gotResults___hundredslimited(self,words,fullResults):
        """will return only three digits as _sofar
        goes together with <__2to9before> and <__1to99after>
        
        """
        didBeforeRules = ['__2to9before']
        doInAfterRules = ['__1to99after', '__1to9after']
    
        print('__hundredslimited, prevRule: %s, nextRule: %s, words: %s'% (self.prevRule, self.nextRule, words))
        lenw = len(words)
        for i in range(lenw):
            if i == 0 and self.prevRule in didBeforeRules:
                pass
            else:
                self._hundreds = 100
            if self.nextRule in doInAfterRules and  i == lenw - 1:
                # leave the "adding" action to nextRule
                pass
            else:
                self._sofar += str(self._hundreds)
                self._hundreds = 0

    def gotResults___thousandslimited(self,words,fullResults):
        """will return only four digits as _sofar
        
        goes together with <__2to9before> and <__1to99after> and possibly <__1to9after>
        
        """
        didBeforeRules = ['__2to9before']
        doInAfterRules = ['__1to99after', '__1to9after']
    
        # print '__hundredslimited, prevRule: %s, nextRule: %s, words: %s'% (self.prevRule, self.nextRule, words)
        lenw = len(words)
        for i in range(lenw):
            if i == 0 and self.prevRule in didBeforeRules:
                pass
            else:
                self._thousands = 1000
            if self.nextRule in doInAfterRules and  i == lenw - 1:
                # leave the "adding" action to nextRule
                pass
            else:
                self._sofar += str(self._thousands)
                self._thousands = 0

    def gotResults___1to99after(self, words, fullResults):
        """should be after __hundredslimited or __thousandslimited
        
        must be defined in doInAfterRules ofter the corresponding rule
        """
        didBeforeRules = ['__hundredslimited', '__thousandslimited']

        print('__1to99after, prevRule: %s, nextRule: %s, words: %s'% (self.prevRule, self.nextRule, words))
        if len(words) > 1:
            raise ValueError("rule __1to99after, expect only one word, got %s: %s"% (len(words), words))
        numWords = self.getNumbersFromSpoken(words)
        num = numWords[0]
        if self.prevRule == '__hundredslimited':
            self._hundreds += num
            self._sofar += str(self._hundreds)
            self._hundreds = 0
        elif self.prevRule == '__thousandslimited':
            self._thousands += num
            self._sofar += str(self._thousands)
            self._thousands = 0
        else:
            print('__1to99after, no valid rule, got: %s, expected one of: %s'% (self.prevRule, didBeforeRules))
            self._sofar += str(num)
            print('_sofar: %s'% self._sofar)

    def gotResults___1to9after(self, words, fullResults):
        """should be after __hundredslimited or __thousandslimited
        
        must be defined in doInAfterRules ofter the corresponding rule
        
        Identical with gotResults___1to99after
        """
        didBeforeRules = ['__hundredslimited', '__thousandslimited']

        print('__1to99after, prevRule: %s, nextRule: %s, words: %s'% (self.prevRule, self.nextRule, words))
        if len(words) > 1:
            raise ValueError("rule __1to99after, expect only one word, got %s: %s"% (len(words), words))
        numWords = self.getNumbersFromSpoken(words)
        num = numWords[0]
        if self.prevRule == '__hundredslimited':
            self._hundreds += num
            self._sofar += str(self._hundreds)
            self._hundreds = 0
        elif self.prevRule == '__thousandslimited':
            self._thousands += num
            self._sofar += str(self._thousands)
            self._thousands = 0
        else:
            print('__1to99after, no valid rule, got: %s, expected one of: %s'% (self.prevRule, didBeforeRules))
            self._sofar += str(num)
            print('_sofar: %s'% self._sofar)


    def gotResults___2to9before(self, words, fullResults):
        """together with __hundredslimited or __thousandslimited
        
        expect one word only, and treat with respect to the next rule
        should be defined in didBeforeRules of the corresponding rule
        """
        # print '__2to9before, prevRule: %s, nextRule: %s, words: %s'% (self.prevRule, self.nextRule, words)
        
        if len(words) > 1:
            raise ValueError("rule __2to9before, expect only one word, got %s: %s"% (len(words), words))
        numWords = self.getNumbersFromSpoken(words, asStr=1)
        numStr = numWords[0]
        num = int(numStr)
        if self.nextRule == '__hundredslimited':
            self._hundreds = num * 100
        if self.nextRule == '__thousandslimited':
            self._thousands = num * 1000
        else:
            print('__2to9before, nextRule is NOT __hundredslimited of __thousandslimited')
            self._sofar += numStr
            print('_sofar: %s'% self._sofar)
    

##8640  8600

#failed to get this working:
#<listtupleargument> exported = (Tuple|List|Argument) (Number <number> | (Variable|String) <dgndictation> | None)+;
    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile()
        self.progInfo = unimacroutils.getProgInfo(moduleInfo)
  
    def initialize(self):
        if not self.language:
            print("no valid language in grammar "+__name__+" grammar not initialized")
            return
        self.load(self.gramSpec)
        # if switching on fillInstanceVariables also fill numbers lists like {n1-9} or {number1to99}
        self.switchOnOrOff() 

    def gotResultsInit(self, words, fullResults):
        # Step 2:
        # initialise the variables you want to collect the numbers in:
        # defining them here makes testing in gotResults easier
        self.number = ''  # for most rules
        self.page = ''    # for page rule
        self.through = '' # only for testnumber2 and pagenumber
        self.pair = ''    # only for the pair rule...
        self.minus = False
        self.ibanHeader = None # for banknumber
        self.ibanCheck =  None # for banknumber
        self.totalNumber = '' # for multiple parts (Amsterdam Zuid, spacingnumber)

        #self.listtupleargument = ''
        #self.Items = [] # for listtupleargumenthallo
        # this line can be used to get the correct fullResults to be used in testIniGrammar.py for
        # for debugging the rules!
        # print 'fullResults of number extended: %s'% fullResults
        
        
    def fillInstanceVariables(self):
        """fills the necessary instance variables

        in this grammar a special rule for Loes Wijers, Amsterdam Zuid.        
        """
        # search commands (for Arnoud):
        self.specialAmsterdamZuid = self.ini.getBool('general', 'special amsterdam zuid') 
        self.filenamePagenumbersPrefix = self.ini.get('general', 'filename page numbers prefix')
        self.filenamePagenumbersPostfix = self.ini.get('general', 'filename page numbers postfix')
        self.lastPage, self.lastThroug = None, None   # filename page numbers
       
    def gotResults_testnumber1(self, words, fullResults):
        # step 3: setting the number to wait for
        # because more numbers can be collected, the previous ones be collected first
        # if you expect only one number, this function can be skipped (but it is safe to call):
        self.collectNumber()
        if self.hasCommon(words, 'number'):
            if self.number:
                # flush previous number
                self.number =self.doMinus('number', 'minus') # convert to int
                self.outputNumber(self.number)
            self.waitForNumber('number')
        else:
            raise ValueError('invalid user input in grammar %s: %s'%(__name__, words))

    def gotResults_testnumber2(self, words, fullResults):
        # step 4 also: if more numbers are expected,
        # you have to collect the previous number before asking for the new
        self.collectNumber()
        # self.minus is not relevant here, as this rule is about integers only...
        # can ask for 'number' or for 'through':
        if self.hasCommon(words, 'number'):
            self.waitForNumber('number')
        elif self.hasCommon(words, 'through'):
            self.waitForNumber('through')
        else:
            raise NumberException('invalid user input in grammar %s: %s'%(__name__, words))

    def gotResults_spacingnumber(self, words, fullResults):
        self.collectNumber()
        self.totalNumber += self.number
        self.waitForNumber('number')

    def gotResults_spacing(self, words, fullResults):
        self.collectNumber()
        self.totalNumber += self.number
        self.waitForNumber('number')
        for w in words:
            spacingchar = self.getFromInifile(w, 'spacing')
            self.totalNumber += spacingchar
        
    def gotResults_pagenumber(self, words, fullResults):
        # step 4 also: if more numbers are expected,
        # you have to collect the previous number before asking for the new
        self.collectNumber()
        # self.minus is not relevant here, as this rule is about integers only...
        # can ask for 'number' or for 'through':
        if self.hasCommon(words, 'page'):
            self.waitForNumber('page')
        elif self.hasCommon(words, 'through'):
            self.waitForNumber('through')
        else:
            raise NumberException('invalid words in pagenumber rule in grammar %s: %s'%(__name__, words))

    def gotResults_filenamelastpage(self, words, fullResults):
        # additional command, compose filename with the last called page number(s).
        # need variables in inifile section [general]: filename pagenumber prefix and filename pagenumber postfix
        # if these are not set, of if no page numbers are "remembered", do nothing
        if self.lastPage:
            if self.lastThroug:
                lp, lt = int(self.lastPage), int(self.lastThroug)
                if lt > lp:
                    pagePart = '%s-%s'% (lp, lt)
                else:
                    pagePart = self.lastPage
            else:
                pagePart = self.lastPage
        else:
            print('numbers extended: no page numbers command issued yet, skip command')
            return
        if not self.filenamePagenumbersPrefix:
            print('%s: command "%s", please specify "filename page numbers prefix" in section [general] of inifile'% (' '.join(words), self.name))
        if not self.filenamePagenumbersPostfix:
            print('%s: command "%s", please specify "filename page numbers postfix" in section [general] of inifile'% (' '.join(words), self.name))
        if self.filenamePagenumbersPrefix and self.filenamePagenumbersPostfix:
            fName = self.filenamePagenumbersPrefix + pagePart + self.filenamePagenumbersPostfix
            action("SCLIP %s"% fName)

    def gotResults_pair(self, words, fullResults):
        # here a bit different logic for the place to collect the previous number.
        #
        self.pair = 'pair'
        if self.hasCommon(words, 'and'):
            self.collectNumber()
            self.number = self.doMinus('number', 'minus')
            self.waitForNumber('pair')
        else:
            self.waitForNumber('number')

    def gotResults_banknumber(self,words,fullResults):
        """get first 8 characters from bank name (list), rest from numbers grammar
        """
        self.ibanHeader = self.getFromInifile(words[-1], 'bankheader')
        self.ibanCheck =  "%.2d"% self.getNumberFromSpoken(words[-2])
        self.waitForNumber('number')

    def gotResults_number(self, words, fullResults):
        # step: when in this rule, the word Minus (or synonym or translation) has been spoken..
        self.minus = True

    def gotResults_numberazone(self, words, fullResults):
        """first part of the number"""
        if not self.totalNumber:
            self.totalNumber = '86'
            # print 'totalNumber: ', self.totalNumber
        self.collectNumber()
        self.totalNumber += self.number
        self.number = ''
        # wait for a number of 6 digits, if not, do NOT output!!
        self.waitForNumber('number', 6)

    def gotResults_numberaztwo(self, words, fullResults):
        """second part of the number"""
        self.collectNumber()
        self.totalNumber += self.number
        self.number = ''
        self.totalNumber += '.'
        # wait for a number of 6 digits, if not, do NOT output!!
        self.waitForNumber('number', 4)

# # # # <numbertwodigits> = {n0-9}{n0-9} | {n10-99};
# # # # <numberthreedigits> = {n0-9}+ | <__hundreds> | {n0-9} <numbertwodigits>;
    def gotResults_numbertwodigits(self, words, fullResults):
        numWords = self.getNumbersFromSpoken(words, asStr=1)
        numberString = ''.join(numWords)
        self._sofar += numberString
        
    def gotResults_numberthreedigits(self, words, fullResults):
        numWords = self.getNumbersFromSpoken(words, asStr=1)
        numberString = ''.join(numWords)
        self._sofar += numberString

    def gotResults_numberfourdigits(self, words, fullResults):
        numWords = self.getNumbersFromSpoken(words, asStr=1)
        numberString = ''.join(numWords)
        self._sofar += numberString

    def gotResults_numbersixdigits(self, words, fullResults):
        numWords = self.getNumbersFromSpoken(words, asStr=1)
        numberString = ''.join(numWords)
        self._sofar += numberString

    def gotResults(self, words, fullResults):
        # step 5, in got results collect the number:
        self.collectNumber()
        
        if self.totalNumber:
            # multiple numbers, AmsterdamZuid
            self.totalNumber += self.number
            print('gotResults totalNumber: ', self.totalNumber)
            if self.message:
                print('Error collecting the number, do NOT output the number...')
                print(self.message)
            else:
                self.outputNumber(self.totalNumber)
            if len(self.totalNumber) == 13 and self.totalNumber[-5] == '.':
                keystroke("{tab}")
            elif len(self.totalNumber) == 5 and self.totalNumber[0] == '.':
                keystroke("{tab}")
            
            return
        
        if self.page:
            ## page numbers rule (for child windows, adobe [acrord32] of pdf24.)
            self.lastPage, self.lastThroug = self.page, self.through
            print('setting lastPage and lastThroug: %s, %s'% (self.lastPage, self.lastThroug))
            isTop = (self.progInfo[2] == 'top')
            ma = getMetaAction('pagestart', progInfo=self.progInfo)
            if not ma:
                print('no metaactions defined for pagestart, stop command %s'% self.progInfo.prog)
                if isTop:
                    if self.through:
                        keystroke(" page %s-%s"% (self.page, self.through))
                    else:
                        keystroke(" page %s"% self.page)
                else:
                    print("cannot finish page command for child window: %s"% repr(self.progInfo))
                return

            action("<<pagestart>>")
            keystroke(self.page)
            if self.progInfo.prog == 'pdf24-creator' and self.through == '':
                self.through = self.page

            if self.through:
                ## only child window here:
                action("<<pagethrough>>")
                keystroke(self.through)
            action("<<pagefinish>>")
            return  # page command
            
        if self.pair:
            self.pair = self.doMinus('pair', 'minus')
            print("(%s, %s) "% (self.number, self.pair))
            
          
        #elif self.listtupleargument:
        #    print 'self.listtupleargument in gotResults: %s'% self.listtupleargument
        #    if self.number:
        #        self.number = self.doMinus('number', 'minus')
        #        self.Items.append(self.number)                    
        #
        #    if self.dictated:
        #        self.Items.append(''.join(self.dictated))
        #        self.dictated = None
        #
        #    result = repr(self.Items)
        #    print 'result: %s'% self.Items
        #    if self.listtupleargument == 'list':
        #        print 'list: %s'% result
        #    elif self.listtupleargument == 'tuple':
        #        result = repr(tuple(self.Items))
        #        print 'tuple: %s'% result
        #    else:
        #        result = repr(tuple(self.Items)).replace(', ', ',')
        #        result = result.replace(', ',',')
        #        print 'argument: %s'% result
        #
        elif self.ibanCheck and self.ibanHeader:
            try:
                # print 'ibanheader: %s, number: %s'% (self.ibanHeader, self.number)
                result = iban.create_iban(self.ibanHeader[:2], self.ibanHeader[2:], self.number)
            except iban.IBANError as err:
                print('invalid iban %s, %s (%s)'% (self.ibanHeader, self.number, err))
                return
            if result[2:4] == str(self.ibanCheck):
                keystroke(result)
            else:
                print('invalid check: %s (%s) '% (self.ibanCheck, result))

        elif self.number:
            # last part if all others failed:
            self.number = self.doMinus('number', 'minus')
            self.outputNumber(self.number)

                      
    def outputNumber(self, number):
        if type(number) in [int, float]:
            number = str(number)

        keystroke(number)
        prog = unimacroutils.getProgName()
        if prog in ['iexplore', 'firefox', 'chrome', 'safari']:
            keystroke('{tab}')
        elif prog in ['natspeak']:
            keystroke('{enter}')
        elif prog in ['excel', 'winword', 'soffice']:
            keystroke('{tab}')



    def doMinus(self, number, minus):
        """return  the minus version of number, is self.minus is set

        pass in the names of the number variable and the name of the minus variable.
        return the wanted number.        
        """
        Nstring = getattr(self, number)
        Nstring = Nstring.strip()
        Minus = getattr(self, minus)
        if not Nstring:
            setattr(self, minus, False)
            return ''

        if Minus:
            if Nstring.startswith('-'):
                Nstring = Nstring[1:]
            else:
                Nstring = '-'+Nstring
            setattr(self, minus, False)
        return Nstring

# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None 

if __name__ == "__main__":
    ## interactive use, for debugging:
    natlink.natConnect()
    try:
        thisGrammar = ThisGrammar(inifile_stem="_number extended")
        thisGrammar.startInifile()
        thisGrammar.initialize()
    finally:
        natlink.natDisconnect()
elif __name__.find('.') == -1:
    # standard startup when Dragon starts:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()




#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#   This script performs some basic tests of the NatLink system.  Dragon
#   NaturallySpeaking should be running with nothing in the editor window
#   (that you want to preserve) before these tests are run.
#   performed.
from pathlib import Path
from pprint import pprint
import sys

from dtactions.unimacro import unimacroactions as actions
import unittest
import TestCaseWithHelpers
import re
import types
import natlink
from dtactions.unimacro import unimacroutils   
from natlinkcore import natlinkutils

from natlinkcore import natlinkutils
import win32gui
reWhiteSpace = re.compile('\s+')

class TestError(Exception):
    pass

class UnimacroTestHelpers(TestCaseWithHelpers.TestCaseWithHelpers):
    """providing helper functions for other unimacro tests
    """
    def squeeze_whitespace(self, text):
        return reWhiteSpace.sub(' ', text)

    def doTestForException(self, exceptionType,command,localVars={}):
        try:
            exec(command, globals(),localVars)
        except exceptionType:
            return
        raise TestError('Expecting an exception to be raised calling '+command)

    def assert_mod_partoftitle(self, expMod, expPartOfTitle=None, text=''):
        """check module and optional part of the window title with the actual

        """        
        mod, title, hndle = natlink.getCurrentModule()
        baseOfMod = natlinkutils.getBaseName(mod).lower()
        self.assert_equal(expMod, baseOfMod, "module unequal " + text)
        if expPartOfTitle:
            self.assert_(title.find(expPartOfTitle) >= 0,
                         'part of title "%s" does not match title "%s" in module "%s" " + text'%
                         (expPartOfTitle, title, mod))

    def doTestActionResult(self, expected, commandString):
        result = Search.doAction(commandString)
        if expected and result:
            pass
        elif not expected and not result:
            pass
        else:
            self.assert_equal(expected, result, "result of action not as expected")
        

    def assert_mod(self, expMod, text=''):
        """check module name with expected

        """        
        mod, title, hndle = natlink.getCurrentModule()
        baseOfMod = natlinkutils.getBaseName(mod).lower()
        self.assert_equal(expMod, baseOfMod, "module unequal " + text)
        
    def failIf_mod(self, expMod, text=''):
        """check module name with expected, should be UNEQUAL

        """
        mod, title, hndle = natlink.getCurrentModule()
        baseOfMod = natlinkutils.getBaseName(mod).lower()
        self.failIf(expMod == baseOfMod, "modules %s and %s are equal, should be unequal "%(expMod, baseOfMod) + text)
                

    #---------------------------------------------------------------------------
    # This utility subroutine will returns the contents of the NatSpeak window as
    # a string.  It works by using playString to select the contents of the
    # window and copy it to the clipboard.  We have to also add the character 'x'
    # to the end of the window to handle the case that the window is empty.

##    def getWindowContents(self):
##        sendkeys('{ctrl+end}x{ctrl+a}{ctrl+c}{ctrl+end}{backspace}')
##        contents = natlink.getClipboard()
##        if contents == 'x':
##            raise TestError,'Failed to read the contents of the NatSpeak window'
##        return contents[:-1]
    def getWindowContents(self):
        action = actions.doAction
        action('CLIPSAVE')
        action("<<selectall>><<copy>>")
        contents = unimacroutils.getClipboard()
        action('CLIPRESTORE')
        return contents
    
    def doTestActionResult(self, expected, commandString):
        result = actions.doAction(commandString)
        if expected and result:
            pass
        elif not expected and not result:
            pass
        else:
            self.assert_equal(expected, result, "result of action not as expected")


    def doTestWindowIsEmpty(self, testName=None):
        contents = self.getWindowContents()
        text = 'the window is NOT EMPTY'
        if testName:
            text += ': ' + testName
        if contents:
            self.fail(text)
            
        

    def doTestWindowContents(self, expected, testName=None):
        contents = self.getWindowContents()
            
        if testName:
            testText = 'Contents of window did not match expected text, testing %s'% testName
        else:
            testText = 'Contents of window did not match expected text'
        if type(expected) == list:
            for e in expected:
                if e == contents: return

            self.fail("contents: %s\nexpected one of: %s\n, %s"% (contents, expected, testText))                
        else:
            self.assert_equal(expected, contents, testText)
            
    def clearWindow(self):
        sendkeys('{ctrl+end}x{ctrl+a}{ctrl+c}{ctrl+end}{backspace}')
    #---------------------------------------------------------------------------
    # Utility function which calls a routine and tests the return value

    def doTestFuncReturn(self, expected,command,localVars={}):
        actual = eval(command,globals(),localVars)
        if actual != expected:
            raise TestError("Function call: %s\n  returned: %s\n  expected %s"%(command,repr(actual),repr(expected)))

    #---------------------------------------------------------------------------
    # This types the keysequence {alt+esc}.  Since this is a sequence trapped
    # by the OS, we must send as system keys.

    def playAltEsc(self):
        sendkeys('{alt+esc}',hook_f_systemkeys)

    #---------------------------------------------------------------------------

    def lookForNatspeakOrDragonPad(self):

        # This should find the NatSpeak window.  If the NatSpeak window is not
        # available (because, for example, NatSpeak was not started before 
        # running this script) then we will get the error:
        #   NatError: Error 62167 executing script execScript (line 1)

        try: natlink.execScript('AppBringUp "NatSpeak"')
        except natlink.NatError:
            raise TestError('The NatSpeak user interface is not running')
        try: natlink.execScript('Start "DragonPad"')
        except natlink.NatError:
            raise TestError('The DragonPad window cannot be started')
        
        # This will make sure that the NatSpeak window is empty.  If the NatSpeak
        # window is not empty we raise an exception to avoid possibily screwing 
        # up the users work.

        if self.getWindowContents():
            raise TestError('The NatSpeak/DragonPad window is not empty')
        mod, title, hndle = natlink.getCurrentModule()
        self.DragonPadMod = mod
        self.DragonPadHndle = hndle
        
    def wait(self, t=None):
        actions.do_Wait(t)

    def killNatspeakOrDragonPad(self):
        unimacroutils.SetForegroundWindow(self.DragonPadHndle)
        mod, title, hndle = natlink.getCurrentModule()
        self.assert_(mod == self.DragonPadMod and hndle == self.DragonPadHndle, "could not get back to Natspeak/DragonPad window")
        actions.killWindow()
        mod, title, hndle = natlink.getCurrentModule()
        self.failIf(mod == self.DragonPadMod and hndle == self.DragonPadHndle, "did not kill Natspeak/DragonPad window")




# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#  _oops.py, can oops the last phrase
#
#  Written by : Quintijn Hoogenboom,QH softwaretraining & advies,
#    July 2002/August 2003
#
"""DragonDictate like "oops" a wrong misunderstood last phrase

The command "Oops" (as was in DragonDictate) is implemented in this
grammar. In the message window the list of alternatives is brought up.
You can choose the item you want. No typing however can be done if the
right choice is not in the list.

This oops command is enhanced with several commands which can be given
in the message window which displays the choice list.
1. Choose # can be followed by the words [middle|strong] to do the
correction and number of times
2. Format #: after that a list of formats is presented from which
the right format can be chosen
3. Delete #: the word is deleted from the active vocabulary

this is a tricky grammar file, because the allResults-flag is
turned on.

 Note: the Oops grammar uses exclusive mode. If my modified
 "natlinkmain.py" is used, this exclusive mode is turned off
 whenever the microphone state with changes.

 

"""
#pylint:disable=R0912, R0914, R0915
#pylint:disable=E1101


import time
import os
from pathlib import Path

import natlink
from natlinkcore import natlinkstatus
from natlinkcore import loader
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj

logHour = -1
logFile = ''

status = natlinkstatus.NatlinkStatus()
natlinkmain = loader.NatlinkMain()

language = status.language
version = status.getDNSVersion()
UDDirectory = status.getUnimacroDataDirectory()
# print 'language: %s (%s)'% (language, type(language))
# print 'version: %s (%s)'% (version, type(version))
# print 'getUnimacroUserDirectory: %s (%s)'% (UUDirectory, type(UUDirectory))
logFolder = Path(UDDirectory)/f'{status.language}_log'/status.user
 
def getLogFileName():
    """get name with date and time, record hour in logHour"""
    #pylint:disable=W0603
    global logHour, logFile
    if not logFolder.exists():
        print(f'Want to create log folder: {logFolder}')
    if not logFolder.exists():
        logFolder.mkdir(parents=True)
    lTime = time.localtime()
    logFile = time.strftime("%Y-%m-%d %H%M", lTime)  + '.txt'
    logHour = lTime[3]
    logFile = os.path.join(logFolder, logFile)
    #print 'logHour: ', logHour, ' log to file: ', logFile

ChooseList = ['1','2','3','4','5','6','7','8','9','10']

# get language for different language versions:
language = unimacroutils.getLanguage()

#  Number of times the correction is executed when choosing weak
#    (default), middle or strong
choiceWeak = 1
choiceMiddle = 5
choiceStrong = 15


#  The following set of formats can be used in the formatting part of
#    this grammar.  Can be enhanced to your own needs.
#  FORMATS and FormatComments MUST MATCH!
FORMATS = {
    # for letters in advance of a variable:
    1: ( unimacroutils.wf_TurnCapitalizationModeOn |
           unimacroutils.wf_TurnOffSpacingBetweenWords |
           unimacroutils.wf_DoNotApplyFormattingToThisWord
          ),
    # for continuing after a variable:
    2: ( unimacroutils.wf_RestoreNormalCapitalization |
            unimacroutils.wf_RestoreNormalSpacing
          ),
    # for continuing after a variable + extra space:
    3:  ( unimacroutils.wf_RestoreNormalCapitalization |
            unimacroutils.wf_RestoreNormalSpacing |
            unimacroutils.wf_AddAnExtraSpaceFollowingThisWord
          ), 
    # for continuing after a variable, no space before:
    4: ( unimacroutils.wf_RestoreNormalCapitalization |
            unimacroutils.wf_RestoreNormalSpacing |
            unimacroutils.wf_NoSpacePreceedingThisWord
          )
    }

if language == 'nld':
    FormatComments = {
        1: 'letters',
        2: 'herstel',
        3: 'herstel + spatie (als "spatie")',
        4: 'herstel + geen spatie ervoor (als ":")'
    }
    ScratchThatCommand = ['schrap', 'dat']
    WordCommand = 'commando'
    WordDictate = 'dictaat'
else:
    FormatComments = {
        1: 'switch on capitalization and switch off spacing (for prefix letters)',
        2: 'restore capitalization and spacing',
        3: 'restore and add a space (like "space-bar")',
        4: 'restore, but no space before (like ":")'
    }
    ScratchThatCommand = ['scratch', 'that']
    WordCommand = 'commando'
    WordDictate = 'dictate'

if len(list(FORMATS.keys())) != len(list(FormatComments.keys())):
    print('warning _oops: FORMATS and FormatComments do not match')
    DoFormatting = 0
else:
    DoFormatting = 1

thisGrammar = None
ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):
    iniIgnoreGrammarLists = ['chooselist'] # are set in this module    
    name = "oops"
    gramSpec = """
<oops> exported = Oops;
# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)

<inoops> exported = Cancel | OK | <choose>;
<choose> = (Choose|Format|Delete|Properties) {chooselist} | (Choose|Format|Delete|Properties) {chooselist}(Weak|Medium|Strong);
<inoops2> exported = Cancel | OK | (Choose Format (1 | 2 | 3| 4));
        """

    def initialize(self):
        # load if ini file gives permission:
        self.switchOnOrOff(activateRule='oops', allResults=1)
        if not self.isLoaded():
            # print(f'{self.name}: grammar is not loaded')
            return
        self.oopsFlag = 0
        self.lastResObj = None
        self.messageHndle = 0
        self.setList('chooselist', ChooseList)
        # if language == 'nld':
        #     print('Grammatica _oops ge-initialiseerd')
        # else:
        #     print('Grammar _oops initialized')
        self.cancelMode()
        self.prevModinfo = None
        if self.logging:
            print(f'{self.name}: logging of utterances to dir:\n\t{logFolder}')
        else:
            print('{self.name}, NO logging of utterances')

    def fillInstanceVariables(self):
        """fills the necessary instance variables

        
        """
        self.logging = self.ini.getBool('general', 'log all utterances')

    def gotBegin(self,moduleInfo):
        """log changed modinfo"""
        if self.checkForChanges:
            self.checkInifile()

        # print('time in gotBegin: %s'% time)
        if self.logging and self.prevModinfo != moduleInfo:
            # print('time: %s'% time)
            lTime = time.localtime()
            if logHour != lTime[3]:
                #print 'get new logfilename'
                getLogFileName()
            self.progInfo = unimacroutils.getProgInfo(modInfo=moduleInfo)
            prog = self.progInfo.prog
            hndle = self.progInfo.hndle
            title = self.progInfo.title
            minutes = time.strftime("%H:%M", time.localtime())

            logToFile(f'\n{minutes}---{prog}---{title}({hndle})')
            self.prevModinfo = moduleInfo
        #else:
        #    print 'no logging or same moduleInfo: %s, %s, %s'% (self.logging, repr(self.prevModinfo), repr(moduleInfo))
  
##        if self.oopsFlag:
##            self.oopsFlag = self.oopsFlag - 1
##            if self.oopsFlag == 0:
##                print 'Canceling oops mode'
##                self.cancelMode()
    def gotResultsInit(self, words, fullResults):
        self.hadChoice = None
        self.nChoice = choiceMiddle  


    def gotResultsObject(self,recogType,resObj):
        #  If in oops mode, no actions are taken in this function.  If
        #    however the grammar was rejected, which is exclusive whenever the
        #    oops mode is on, the warning message is printed.hello
        if self.logging:
            if natlink.getCallbackDepth() < 3:
                try:
                    words = resObj.getWords(0)
                except (natlink.OutOfRange, IndexError):
                    words = "<???>"

                resCode = -1
                isDictate = -1
                k, rest = -1, -1
                for i in range(10):
                    try:
                        res = resObj.getResults(i)
                        ##TODO DOUG
                        # try:
                        #     lenWave = len(resObj.getWave())
                        # except natlink.DataMissing:
                        #     lenWave = 0
                        lenWave = 0
                        k,rest = int(lenWave/1000), lenWave%1000
                        resCode = res[0][1]
                        resCode = resCode & 0x7fffffff
                        if i == 0:
                            isDictate = not resCode
                        if resCode:
                            break
                    except (natlink.OutOfRange, IndexError):
                        break

                if isDictate == -1:
                    line = '  <???> '
                elif isDictate and not resCode:
                    line = f'  {" ".join(words)} (only dictate alternatives)'
                elif isDictate:
                    line = f'  {" ".join(words)} (dictate, command alternatives)'
                else:
                    line = f'  {words} (command {resCode})'
                if k == -1:
                    logToFile(line+'(nowave)')
                elif rest <= 500:
                    logToFile(line+f'({k}k+{rest})')
                else:
                    logToFile(line+'({k+1}k-{1000-rest})')
                    
            
        if recogType == 'other':
            # remember result for repeat grammar:
            if natlink.getCallbackDepth() < 3:
                self.lastResObj = resObj
##            else:
##                print 'callbackdepth" %s, words: %s'%(natlink.getCallbackDepth(), resObj.getWords(0))
                

    def gotResults_oops(self,words,fullResults):
##        t0 = time.time()
##        try:fellow,hellohellohellohello hellohello,fellow hello hello hello
##        try:
##            unimacroutils.SetForegroundWindow(self.messageHndle)
##        except:
        unimacroutils.switchToWindowWithTitle("Messages from Python Macros")
##            self.messageHndle = natlink.getCurrentModule()[2]
##            print 'new handle for message window: %s'% self.messageHndle
##        print 'time to switch:', time.time() - t0
        if not self.lastResObj:
            print('no object to Oops')
            unimacroutils.Wait(0.1)
            unimacroutils.returnFromMessagesWindow()
            return
        #  Go through the alternatives in the results object
        SingleWord = 0  # so formatting can be done
        self.FirstIsDictate = 0
        #waveText = self.lastResObj.getWave()
        #print 'gesproken: %s' % len(waveText)
        for i in range(10):
            try:
                res = self.lastResObj.getResults(i)
                words = self.lastResObj.getWords(i)
                resCode = res[0][1]
                resCode = resCode & 0x7fffffff
                wordInfo = self.lastResObj.getWordInfo(i)
##                pron = []
##                if wordInfo:
##                    pron = map(lambda x: x[6].join(wordInfo))
                if resCode:
                    print(f'{i+1}:\t[{words}]  ({WordCommand} {resCode})')
                else:
                    print(f'{i+1}:\t{words}  ({WordDictate})')
                    if i == 0:
                        self.FirstIsDictate = 1
                    if len(words) == 1:
                        SingleWord = 1

                    
            except (natlink.OutOfRange, IndexError):
                break
        else:
            i = 10
        if language == 'nld':
            print(f'OK, Annuleren, of Kies 1, ..., Kies {i} [middel|sterk]')
            if SingleWord:
                print(f'of Format #, Verwijder # of Eigenschappen #  (# = 1, ..., {i}')
        else:
            print(f'OK, Cancel, or Choose 1, ..., Choose {i} [Medium|Strong]')
            if SingleWord:
                print(f'or Format #, Delete # or Properties # (# = 1, ..., {i})')

        self.oopsFlag = 3
        self.activateSet(['inoops'],exclusive=1)
        self.setList('chooselist', ChooseList[:i])
        unimacroutils.switchToWindowWithTitle("Messages from Python Macros")

    def gotResults_inoops(self,words,fullResults):
        # nCorr = choiceWeak
        # choice = 0
        if not self.lastResObj:
            self.cancelMode()
            unimacroutils.returnFromMessagesWindow()
            
        if self.hasCommon(words, 'Cancel'):
            texts = dict(nld='oeps geannuleerd')
            t = texts.get(self.language, 'oops canceled')
            self.DisplayMessage(t)
            print('cancelling exclusive oops mode')
            unimacroutils.Wait()
            self.cancelMode()
            unimacroutils.returnFromMessagesWindow()
            return
        choice = 0
        if self.hasCommon(words, 'OK'):
            print('OK')
            self.hadChoice = 1
            #ww = []
            
    def gotResults_choose(self, words, fullResults):
        self.nChoice = None
        if self.hasCommon(words[-1], 'Medium'):
            self.nChoice = choiceMiddle
        elif self.hasCommon(words[-1], 'Strong'):
            self.nChoice = choiceStrong
        elif self.hasCommon(words[-1], 'Weak'):
            self.nChoice = choiceWeak
        if self.nChoice:
            del words[-1]
        else:
            self.nChoice = choiceMiddle


        if words[-1] in ChooseList:
            choice = int(words[-1])
        if not choice:
            print('no valid choice given')
            unimacroutils.Wait(0.2)
            self.cancelMode()
            unimacroutils.returnFromMessagesWindow()
            return
        try:
            newWords = self.lastResObj.getWords(choice-1)
        except natlink.OutOfRange:
            i = choice-1
            print(f'_oops, choose {i}, no result number {i}')
            return
        res = self.lastResObj.getResults(choice-1)
        resCode = res[0][1]
        resCode = resCode & 0x7fffffff
        # formatting:===========================================
        if 'Format' in words:
            if not DoFormatting:
                print('formatting options invalid!')
                return
            
            if resCode:
                print('no formatting can be done on a command!')
                time.sleep(1.5)
            elif len(newWords) > 1:
                print('no formatting can be done on a list of words')
                time.sleep(1.5)
            else:
                self.newWord = newWords[0]
                fKeys = list(FORMATS.keys())
                fKeys.sort()
                fcKeys = list(FormatComments.keys())
                fcKeys.sort()
                if fKeys != fcKeys:
                    print('keys of FORMATS and FormatComments do not match')
                    return
                numChoices = len(fKeys)
                if language == 'nld':
                    print(f'Formatteren van: {self.newWord}')
                    print(f'Kies Format 1, ..., {numChoices} of zeg "Annuleren"')
                elif language == 'enx':
                    print(f'Formating: {self.newWord}')
                    print(f'Choose Format 1, ..., {numChoices}, or say "Cancel"')
                else:
                    print('invalid language, skip this')
                    self.cancelMode()
                    return
                for n in range(numChoices):
                    print(f'{n+1}:\t{FormatComments[n+1]}')
                    
                #  Entered the new exclusive grammar rules, for the right
                #    format to be chosen
                self.oopsFlag = 3
                self.activateSet(['inoops2'],exclusive=1)
                return
        # deleting:===========================================
        elif self.hasCommon(words, ['Delete',  'Verwijder']):
            if resCode:
                print('no delete of a command!')
                time.sleep(1.5)
            elif len(newWords) > 1:
                print('no delete on a list of words')
                time.sleep(1.5)
            else:
                natlink.deleteWord(newWords[0])
                print(f'deleted: {newWords[0]}')
        elif self.hasCommon(words, ['Properties','Eigenschappen']):
            if resCode:
                print('no properties on a command!')
                time.sleep(1.0)
            elif len(newWords) > 1:
                print('no properties of a list of words')
                time.sleep(1.0)
            else:
                self.newWord = newWords[0]
                props = natlink.getWordInfo(self.newWord)
                print(f'properties of {self.newWord}: {props}')
                p = unimacroutils.ListOfProperties(props)
                if p:
                    for pp in p:
                        print(pp)
                    time.sleep(4.0)
        elif self.hasCommon(words, ['Choose', 'Kies', 'OK']):
            hadChoose = 1
            print(f'correcting: {newWords} ({self.nChoice})')
            for i in range(self.nChoice):
                result = self.lastResObj.correction(newWords)
                if not result:
                    print('correction failed')
                    break
            else:
                print(f'corrected {self.nChoice} times')
        else:
            print(f'invalid word in command: {words}')
            time.sleep(2.0)
        time.sleep(1.0)
        self.cancelMode()
        unimacroutils.returnFromMessagesWindow()
        #  Like in DragonDictate, when the word was not a command but a
        #    dictate word, the last phrase is scratched and replaced by the new
        #    text or the new command.
        if hadChoose and self.FirstIsDictate:
            natlink.recognitionMimic(ScratchThatCommand)
            natlink.recognitionMimic(newWords)

    def gotResults_inoops2(self,words,fullResults):
        if self.lastResObj:
            fstring = ''
            if self.hasCommon(words, 'Cancel'):
                unimacroutils.Wait()
                self.cancelMode()
                unimacroutils.returnFromMessagesWindow()
                return
            try:
                fNum = int(words[-1])
            except ValueError:
                fNum = ''
            
            fstring = FORMATS.get(fNum, '')

            if fstring and self.newWord:
                oldFormat = natlink.getWordInfo(self.newWord)
                if fstring == oldFormat:
                    print(f'format of {self.newWord} is already: {fstring}')
                else:
                    ##TODO QH
                    print(f'formatting word: {self.newWord} from hex {oldFormat:x} to hex: {fstring} (cannot format string to hex)'%(self.newWord, oldFormat, fstring))
                    natlink.setWordInfo(self.newWord, fstring)
            self.newWord = ""
        time.sleep(1.0)
        self.cancelMode()
        unimacroutils.returnFromMessagesWindow()

    def cancelMode(self):
        #print "end of oops exclusive mode", also called when microphone is toggled.
##        print 'resetting oops grammar'
        # isActive = self.isActive()
        if self.isExclusive():
            print('oops, cancel exclusive mode')
            self.activateSet(['oops'], exclusive = 0)
        self.newWord = ''

def logToFile(line):
    """logging !
    """
    if not logFile:
        getLogFileName()
    if not logFile:
        print('_oops, logging: cannot find valid name for logFile')
        return
    with open(logFile, 'a', encoding='utf8') as fp:
        if not line.endswith('\n'):
            line += '\n'
        fp.write(line)

# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None

if __name__ == "__main__":
    ## interactive use, for debugging:
    natlink.natConnect()
    try:
        thisGrammar = ThisGrammar()
        thisGrammar.initialize()
    finally:
        natlink.natDisconnect()
elif __name__.find('.') == -1:
    # standard startup when Dragon starts:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()
    natlinkmain.set_on_mic_off_callback(thisGrammar.cancelMode)



# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
# 
# Author: Bart Jan van Os, Version: 0.1
"""This file implements a crude Training dialog to Train phrases

"""

import string,sys,pickle

import win32ui,win32con,win32api,commctrl
from pywin.mfc import dialog
from pywin.framework import dlgappcore

import natlink
from natlinkcore.natlinkutils import *
from natlinkutilsbj import SetMic


# The training dialog gives a list at the bottum with each phrase to train. After training
# is started, each phrase to train is displayed at the top, any training results
# are displayed in the control in between. If the recognized results are different from
# the phrase to train, the phrase can be trained again until
MaxRepeat=3
# is reached. If the result is correct, the next phrase will be selected.

# You can stop training at any time by 'stop training'. This stops the training process, but leaves results
# for later batch processing. Switching to another window and saying something has the same result.
# Cancel will quit the dialog without batchprocessing. NOTE: any 'results obj.correction'
# changes are processed immediately, before possible quitting.
# Choosing Done will start processing the results in the batch training mode, and leave the dialog. 
# You can use ' skip phrase' or ' previous phrase' to go forth and back through the phrases to train,
# as wel as use the mouse/keyboard in the list at the bottom.

## Be careful with using this training process.  Because it uses the batch training
## mode,(I presume) it also affects the acoustic models. One may question whether
## the words to be trained are a good representation of your use of the system,
## and you have to be sure that the words are in your vocabulary.
## In the end, general training is probably much better for changing the acoustic
## models. However, it has helped me to train a lot of
## often misrecognized commands without having to type them.  I always do some
## general training afterwards.  Possibly, training mode should be changed to
## only use the obj.correction method.
## This can be accomplished by setting the next constant to zero.
TrainInBatchMode=1


# Things to do:
# - add a control for selecting the training mode.define method
# - add a control for selecting the maximum repeat count.

class TrainDialogGrammar(GrammarBase):


    gramSpec = """
        <dgndictation> imported;        #This must be the first rule      
        <TrainedWord> = {TrainedWord};  #This must be the second rule      
        <ChangePhrase> = previous phrase | skip phrase;
        <Buttons> = (start [training]| (stop|cancel) [training | this] | done );
        <TrainButtons> = click stop | click done | click cancel | ((stop |cancel) (training | this));
        <normalState> exported =  <Buttons>;
        <trainingState> exported =   <TrainButtons> | <ChangePhrase> |
                                     <TrainedWord> | <dgndictation>;
    """

    def initialize(self,dlg):
        self.dlg = dlg
        self.load(self.gramSpec,allResults=1)
        self.thisWindow=dlg.GetSafeHwnd()
        self.emptyList('TrainedWord')

    def terminate(self):
        self.dlg = None
        self.unload()

    def setMode(self):
        if not self.dlg.AreTraining:
            #self.activateSet(['normalState'],window=self.thisWindow,exclusive=0)
            self.activateSet(['normalState'],exclusive=0)
        else:
            self.activateSet(['trainingState'],exclusive=1)

    def gotBegin(self, moduleInfo):
        winHandle = matchWindow( moduleInfo, 'pythonw', 'Train Phrases' )
        if winHandle:
        #if winHandle==self.thisWindow:
            self.setMode()
        else:
            self.deactivateAll()
            if self.dlg.AreTraining: self.dlg.onStop(0,0)


# This processes the results within the training mode
    def gotResultsTraining(self,recogType,resObj):
        if recogType == 'reject':
            print('rejected')
        else:
            results= convertResults(resObj.getResults(0))
            rule=list(results.keys())[0]
            #print results.keys()
            if (rule<=2):
                self.dlg.trainedPhrases.append((resObj,self.dlg.phrase))
                results=' '.join(resObj.getWords(0))
                print(results)

                #go next or repeat
                lowResults=results.lower()
                lowSpokenForm=lowResults.split('\\')[-1]
                lowPhrase=self.dlg.phrase.lower()
                if (lowResults==lowPhrase) or (lowSpokenForm==lowPhrase):
                    self.dlg.nextWord()
                else:
                    self.dlg.Repeat=self.dlg.Repeat+1
                    if self.dlg.Repeat>MaxRepeat:
                        self.dlg.nextWord()
        

    # this callback is where we get the results object
    def gotResultsObject(self,recogType,resObj):
        if self.dlg.AreTraining:
            self.gotResultsTraining(recogType,resObj)

    def gotResults_Buttons(self,words,fullResults):
        if 'done' in words:
            key='{Alt+d}'
        elif 'start' in words:
            key='{Alt+s}'
        elif 'cancel' in words:
            key='{Alt+c}'
        elif 'stop' in words:
            key='{Alt+p}'
        sendkeys(key)

    def gotResults_TrainButtons(self,words,fullResults):
        self.gotResults_Buttons(words,fullResults)

    def gotResults_ChangePhrase(self,words,fullResults):
        if 'previous' in words:
            key='{Up}'
        elif 'skip' in words:
            key='{Down}'
        sendkeys(key)



IDC_EDIT=1000
IDC_PHRASE=1010
IDC_START=1011
IDC_STOP=1012


class TrainDialog(dialog.Dialog):
#    def __init__(self, title, hierList, bitmapID = win32ui.IDB_HIERFOLDERS, dlgID = win32ui.IDD_TREE, dll = None, childListBoxID = win32ui.IDC_LIST1):

    def __init__ (self, title, phrases, MicStart):
        #self.iconId = win32ui.IDR_MAINFRAME
        dialog.Dialog.__init__ (self, self._maketemplate(title))
        self.colHeadings=['Phrase']
        self.items = phrases
        self.MicStart=MicStart
        self.AreTraining=0
        self.CurItem=0
        self.phrase=''
        self.Repeat=0
        self.Batch=0

    def _maketemplate(self, title):
        style = (win32con.WS_DLGFRAME | 
                win32con.WS_SYSMENU | win32con.WS_VISIBLE )
        ls = (
            win32con.WS_CHILD           |
            win32con.WS_VISIBLE         |
            commctrl.LVS_ALIGNLEFT      |
            commctrl.LVS_REPORT
            )
        bs = (
            win32con.WS_CHILD           |
            win32con.WS_VISIBLE
             )
        child = win32con.WS_CHILD | win32con.WS_VISIBLE
        return [ [title, (0, 0, 200, 200), style, None, (8, "MS Sans Serif")],
            ["SysListView32", None, win32ui.IDC_LIST1, (0, 0, 450, 200), ls],
            [128,   "&Start", IDC_START, (10, 0, 50, 14), bs | win32con.BS_DEFPUSHBUTTON],
            [128,   "Sto&p", IDC_STOP, (10, 0, 50, 14), bs ],                 
            [128,   "&Done", win32con.IDOK, (10, 0, 50, 14), bs],
            [128,   "&Cancel",win32con.IDCANCEL,(0, 0, 450, 200), bs],
            [129,"",IDC_EDIT,(0,0,0,0),child|win32con.ES_MULTILINE|win32con.WS_BORDER|
                    win32con.ES_AUTOVSCROLL|win32con.WS_VSCROLL],
            [130,"",IDC_PHRASE,(7,60,80,8),child|win32con.SS_LEFT],                 
            ]

    def OnInitDialog (self):
        rc = dialog.Dialog.OnInitDialog (self)
        self.HookMessage (self.on_size, win32con.WM_SIZE)
        self.HookNotify(self.OnListItemChange, commctrl.LVN_ITEMCHANGED)
        self.HookCommand(self.onStart, IDC_START)
        self.HookCommand(self.onStop, IDC_STOP)        
        self.itemsControl = self.GetDlgItem(win32ui.IDC_LIST1)
        self.butOK = self.GetDlgItem(win32con.IDOK)
        self.butCancel = self.GetDlgItem(win32con.IDCANCEL)
        self.butStart = self.GetDlgItem(IDC_START)
        self.butStop = self.GetDlgItem(IDC_STOP)        
        self.Output= self.GetDlgItem(IDC_EDIT)
        self.Phrase= self.GetDlgItem(IDC_PHRASE)
        self.SelItems=[]

        self.FillList()

        size = self.GetWindowRect()
        self.LayoutControls(size[2]-size[0], size[3]-size[1])
        self.butOK.EnableWindow(0) # wait for first selection
        self.butStop.EnableWindow(0) # wait for first selection
        self.SetFocus()
        #Select First Item
        self.itemsControl.SetItemState(0,commctrl.LVNI_SELECTED+commctrl.LVNI_FOCUSED,255)

        self.oldStdout = sys.stdout        
        sys.stdout = self
        self.message = ''
        self.grammar = TrainDialogGrammar()
        self.grammar.initialize(self)
        self.trainedPhrases=[]        
        natlink.setChangeCallback(self.changeCallback)
        SetMic(self.MicStart)        
        return rc

    def changeCallback(self,what,param):
        if what == 'user':
            print('User changed.  New user is', param[0])
            self.onStop(0,0)
        elif what == 'mic':
            print('Microphone is ', param)
            if param=='off':
                self.onStop(0,0)

    def OnDestroy(self,msg):
        sys.stdout = self.oldStdout
    
    def write(self,text):
        self.message = self.message + text.replace('\n','\r\n')
        self.SetDlgItemText(IDC_EDIT,self.message)
        self.GetDlgItem(IDC_EDIT).SendMessage(win32con.EM_SETSEL,0x7FFF,0x7FFF)
        self.GetDlgItem(IDC_EDIT).SendMessage(win32con.EM_SCROLLCARET,0,0)

    def FillList(self):
        size = self.GetWindowRect()
        width = size[2] - size[0] - (10)
        itemDetails = (commctrl.LVCFMT_LEFT, width, "Phrase", 0)
        self.itemsControl.InsertColumn(0, itemDetails)
        index = 0
        for item in self.items:
            index = self.itemsControl.InsertItem(index+1, str(item), 0)

    def SetPhraseToTrain(self):
        self.Repeat=1
        if self.AreTraining:
            self.phrase=self.items[self.CurItem]
            self.SetDlgItemText(IDC_PHRASE,self.phrase.center(90))
            self.grammar.setList('TrainedWord',[self.phrase])
        else:
            self.SetDlgItemText(IDC_PHRASE,'')
            #self.setList('TrainedWord',[])            
                

    def OnListItemChange(self,std, extra):
        (hwndFrom, idFrom, code), (itemNotify, sub, newState, oldState, change, point, lparam) = std, extra
        oldSel = (oldState & commctrl.LVIS_SELECTED)!=0
        newSel = (newState & commctrl.LVIS_SELECTED)!=0
        if oldSel != newSel:
            try:
                if newSel:
                    self.CurItem=itemNotify
                    self.SetPhraseToTrain()
            except win32ui.error:
                pass


    def LayoutControls(self, w, h):
        self.Phrase.MoveWindow((0,10,w,27))                
        self.Output.MoveWindow((0,30,w,97))                
        self.itemsControl.MoveWindow((0,100,w,h-30))
        self.butStart.MoveWindow((10, h-24, 60, h-4))
        self.butStop.MoveWindow((70, h-24, 120, h-4))                
        self.butCancel.MoveWindow((w-60, h-24, w-10, h-4))
        self.butOK.MoveWindow((w-120, h-24, w-70, h-4))


    def on_size (self, params):
        lparam = params[3]
        w = win32api.LOWORD(lparam)
        h = win32api.HIWORD(lparam)
        self.LayoutControls(w, h)

                        
    def onStart(self,nID,code):
        print('training started')
        self.AreTraining=1
        if TrainInBatchMode:
            if not self.Batch:
                natlink.startTraining('batchadapt')
                self.Batch=1
                print('batch training started')
            else:
                print('wat')
        self.butOK.EnableWindow(1) 
        self.butStop.EnableWindow(1) 
        self.butStart.EnableWindow(0)
        self.CurItem = self.CurItem - 1
        self.itemsControl.SetFocus()
        self.nextWord()

    def onStop(self,nID,code):
        print('training stopped')
        self.AreTraining=0
        self.butStop.EnableWindow(0) 
        self.butStart.EnableWindow(1)
        self.SetPhraseToTrain()
        self.itemsControl.SetFocus()

    def correctResults(self):
        if len(self.trainedPhrases)>0:
            print('correcting results')
            for (resObj,phrase) in self.trainedPhrases:
                results=' '.join(resObj.getWords(0))
                if not resObj.correction(phrase):
                    if resObj.correction(phrase.lower()):
                        print('correcting %s by lowercase %s.' % (results,phrase))
                    else:
                        print('correcting %s by %s rejected.' % (results,phrase))
            self.trainedPhrases=[]        
        

    def OnOK(self):
        self.onStop(0,0)
        self.correctResults()
        if TrainInBatchMode and self.Batch:
            print('batch processing results')
            try:
                natlink.finishTraining()
                print('results processed')
            except:
                print('results processing not possible')
            self.Batch=0
        else:
            print('Done')
        #self._obj_.OnOK()


    def EndDialog(self, rc):
        self.trainedPhrases=[]
        if TrainInBatchMode and self.Batch:       
            natlink.finishTraining(0)
            print('results not processed')
        self._obj_.EndDialog(rc)
        return


    def nextWord(self):
        if self.CurItem < len(self.items)-1:
            self.itemsControl.SetFocus()
            self.itemsControl.SetItemState(self.CurItem,0,commctrl.LVNI_SELECTED+commctrl.LVNI_FOCUSED)            
            self.CurItem = self.CurItem + 1
            self.itemsControl.SetItemState(self.CurItem,0,commctrl.LVNI_SELECTED+commctrl.LVNI_FOCUSED)                        
            self.itemsControl.SetItemState(self.CurItem,commctrl.LVNI_SELECTED+commctrl.LVNI_FOCUSED,255)
            self.itemsControl.EnsureVisible(self.CurItem, 0)
        else:
            self.onStop(0,0)
  


def demodlg ():
    Title='Train Phrases'
    p=[('word 1'),('word 2'),('word 3')]
    dlg=TrainDialog(Title, p, 1)
    try:
        natlink.natConnect(1)
    except:
        return
    dlg.DoModal()
    natlink.natDisconnect()    


if __name__=='__main__':
    #pass
    demodlg()
    #demomodeless()




#probably not used anymore (Quintijn)
import os.path

def convertTo83Path(fullPath,removeSpacesOnly=0):
    """tries to convert (valid) long path names to Dos83 pathnames

    """
    if fullPath.find('-') >= 0:
        print('found "-", dos83 ignore: %s'% fullPath)
        return fullPath
    # only possible if fullPath exists.
    # no garantee that the converted path refers to the same location
    # I have not found standard functions for this conversion,
    # nor do I know how to compare equivalence: os.path.samefile is not available in Win
    # Are there any winApi calls?
    if not os.path.exists(fullPath):
        print('Path conversion failed. This is not a valid path:')
        print(fullPath)
        return ''
    # go try convert
    (name,ext)=os.path.splitext (fullPath)
    splitPath=name.split('\\')
    splitPath83=splitPath[:]
    for subPath in splitPath:
        concatPath=''.join(subPath.split())
        if (concatPath!=subPath) or ((not removeSpacesOnly) and (len(subPath)>8)):
            # convert subpath to Dos 8.3 convention
            concatPath=concatPath[0:min(6,len(concatPath))]
            i=splitPath.index(subPath)
            altConcatPath=[]
            for n in range(1,10): # try simple rule and find those that exist
                splitPath83[i]=concatPath+'~'+str(n)
                Path83=' '.join(splitPath83,'\\')+ext
                if os.path.exists(Path83):
                    if os.stat(Path83)==os.stat(fullPath):
                        altConcatPath.append(concatPath+'~'+str(n))
            # how many existing with matching stats did we find?        
            if len(altConcatPath)==1:
                splitPath83[i]=altConcatPath[0]
            elif len(altConcatPath)>1:
                splitPath83[i]=altConcatPath[0]
                print('Warning. Non unique Path conversion.')
            else:
                splitPath83[i]=concatPath+'??'
                print('Warning. Insufficient rule for conversion.')
    Path83=' '.join(splitPath83,'\\')+ext
    if not os.path.exists(Path83):
        print('Path conversion failed. This is not a valid Dos 8.3 path:')
        print(Path83)
        return ''
    else:    
        return Path83

def convertSpacedSubPaths(fullPath):
    return convertTo83Path(fullPath,removeSpacesOnly=1)

print(convertTo83Path('C:\Program Files'))#print convertTo83Path('C:\Program Files\Common Files\Microsoft Shared\MSCREATE.DIR')#print convertTo83Path('C:\Program Files\ArrayVisualizer\SAMPLES\SAMPLES.HTM')#print convertSpacedSubPaths('C:\Program Files\ArrayVisualizer\SAMPLES\SAMPLES.HTM')


# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory
#
#  _brackets.py
# written by: Quintijn Hoogenboom (QH softwaretraining & advies)
# august 2003, revision october 2011, python3: september 2021
""" unimacro grammar that puts brackets, braces etc.

Basically two ways are implemented:

1. Say: "between brackets hello comma this is a test"
2. select text, and say "between brackets"
3. say "empty braces"
4. say "here parens"

Notes:
======
1. If you start or end with space-bar (or other white space),
   this will be put OUTSIDE the brackets.

2. No capitalisation is done unless you call as directive.

3. Dictation errors cannot be corrected with the spell window. Select, 
dictate again and then correct then if needed.

Note: the natlinkclipboard module from dtactions is not ready for use. Use
the unimacroutils module of unimacro.


"""
#pylint:disable=C0115, C0116, W0201, W0613
from natlinkcore import nsformat
from dtactions.unimacro import unimacroutils
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doKeystroke as keystroke
# from dtactions.natlinkclipboard import Clipboard
import unimacro.natlinkutilsbj as natbj
import natlink

language = unimacroutils.getLanguage()

ancestor = natbj.DocstringGrammar
class BracketsGrammar(ancestor):
    language = unimacroutils.getLanguage()
    name = "brackets"

    def initialize(self):
        if not self.language:
            print("no valid language in grammar "+__name__+" grammar not initialized")
            return
        self.load(self.gramSpec)
        self.switchOnOrOff()

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile()
        if self.mayBeSwitchedOn == 'exclusive':
            self.switchOnOrOff()

    def gotResultsInit(self, words, fullResults):
        self.dictated = ''  # analysis of dgndictation or dgnletters
        self.pleft = self.pright = '' # the left and right parts of the brackets
        self.here, self.between, self.empty = False, False, False
        if self.mayBeSwitchedOn == 'exclusive':
            print(f'recog brackets, switch off mic: {words}')
            natbj.SetMic('off')

    def importedrule_dgndictation(self, words):
        #do with nsformat functions:
        self.dictated, dummy = nsformat.formatWords(words, state=-1)  # no capping, no spacing
        #print '-result of nsformat: |%s|'% repr(self.dictated)

    def rule_brackets(self, words):
        "<before> {brackets}+ [<dgndictation>]"
        for w in words:

            p = self.getFromInifile(w, 'brackets')
            if not p:
                print(f'no valid brackets found for word: "{w}"')
                continue
            #print 'brackets, found: %s, %s'% (w, p)
            if len(p) > 2 and p.find("|") > 0:
                pList = p.split("|")
                newpleft = pList[0]
                newpright = pList[1]
            else:
                lenph = len(p)//2
                newpleft, newpright = p[:lenph], p[lenph:]
            # make more brackets together, from outer to inner:
            self.pleft = self.pleft + newpleft
            self.pright = newpright + self.pright
        # print(f'result rule_brackets: |{self.pleft}|, pright: |{self.pright}|')

    def subrule_before(self, words):
        "(here|between|empty)+"

        for w in words:
            if self.hasCommon(w, 'between'):   # this is the trigger word, ignore
                self.between = True
            if self.hasCommon(w, 'here'):   # this is the trigger word, ignore
                self.here = True
            if self.hasCommon(w, 'empty'):   # this is the trigger word, ignore
                self.empty = True
  

    def gotResults(self, words, fullResults):

        #  see if something selected, leaving the clipboard intact
        #  keystroke('{ctrl+x}')  # try to cut the selection
        # if no text is dictated, self.dictated = ""
        text, leftTextDict, rightTextDict = stripFromBothSides(self.dictated)

        if self.here:
            print('do a left buttonClick')
            unimacroutils.buttonClick('left', 1)
            unimacroutils.visibleWait()


        if self.empty:
            if self.dictated:
                print(f'_brackets, warning, dictated text "{self.dictated}" is ignored, because of keyword "empty"')
            self.do_keystrokes_brackets()
            return
        
        # only if no dictated text, try to cut the selection (if there, add one char for safety with
        # the clipboard actions, only fails when at end of file)
        if not self.dictated:
            unimacroutils.saveClipboard()
            keystroke('{shift+right}')   # take one extra char for the clipboard to hit
            action('<<cut>>')
            action('W')
            cb_text = unimacroutils.getClipboard()
            unimacroutils.restoreClipboard()
            if cb_text:
                text, lastchar = cb_text[:-1], cb_text[-1]
            else:
                action('<<undo>>')
                raise OSError('no value in clipboard, restore cut text')
            print(f'_brackets, got from clipboard: "{text}" + extra char: "{lastchar}"')
            self.do_keystrokes_brackets(text=text, lastchar=lastchar)
            return

        self.do_keystrokes_brackets(text=text, l_spacing=leftTextDict, r_spacing=rightTextDict)

#
    def do_keystrokes_brackets(self, text='', lastchar='', l_spacing='', r_spacing=''):
        """do the pleft text pright keystrokes with spacing issues
        
        handle the "between" keyword here!
        """
        keystroke(l_spacing)
        keystroke(self.pleft)
        unimacroutils.visibleWait()
        if text:
            keystroke(text)
            unimacroutils.visibleWait()
        keystroke(self.pright)
        keystroke(r_spacing)
        if lastchar:
            keystroke(lastchar)
            keystroke("{left %s}"% len(lastchar))
        if self.between:
            keystroke("{left %s}"% len(self.pright))
        

def stripFromBothSides(text):
    """strip whitespace from left side and from right side and return
the three parts

    input: text
    output: stripped, leftSpacing, rightSpacing
    """
    if not text:
        return "", "", ""
    leftText = rightText = ""
    lSpaces = len(text) - len(text.lstrip())
    leftText = rightText = ""
    if lSpaces:
        leftText = text[:lSpaces]
    text = text.lstrip()
    rSpaces = len(text) - len(text.rstrip())
    if rSpaces:
        rightText = text[-rSpaces:]
    text = text.rstrip()
    return text, leftText, rightText

# standard stuff Joel (adapted in course of time, QH)
def unload():
    #pylint:disable=W0603, E0601
    global bracketsGrammar
    if bracketsGrammar:
        bracketsGrammar.unload()
    bracketsGrammar = None


if __name__ == "__main__":
    natlink.natConnect()
    try:
        bracketsGrammar = BracketsGrammar(inifile_stem='_brackets')
        bracketsGrammar.startInifile()
        bracketsGrammar.initialize()
    finally:
        natlink.natDisconnect()
else:
    bracketsGrammar = BracketsGrammar()
    if bracketsGrammar.gramSpec:
        bracketsGrammar.initialize()
    else:
        bracketsGrammar = None
        



"""unittest of unimacro stuff

This module tests unimacro things.

The test files (unittest) in the subfolder "unimacro_test", and have name like "BasicTest.py".

The test class name MUST be identical to the filename.

Tests can be run one alone (call unimacro test testname) or
or all (call unimacro test all)

instance variables that are passed to all tests:
-self.doAll (so the test knows it is a single test or a complete suite)

"""
import os
import sys
import unittest
import glob
import natlink
from natlinkcore import natlinkstatus
import unimacro.natlinkutilsbj as natbj
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import utilsqh
from dtactions.unimacro import unimacroactions as actions

status = natlinkstatus.NatlinkStatus()

class UnittestGrammar(natbj.IniGrammar):
    language = unimacroutils.getLanguage()        
    name = 'unimacro test'
    iniIgnoreGrammarLists = ['tests'] # are set in this module
    gramSpec = """
<onetest> exported = unimacro test {tests};
<alltests> exported = unimacro test (all);
    """

    def initialize(self):
        self.load(self.gramSpec)
        self.activateAll()
        self.switchOnOrOff()
        self.setList('tests', self.getTestNames()) # also fill self.testNames, self.allTests
        self.filesToSkipInResult = ["unittest.py", "TestCaseWithHelpers.py"]

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile()
            self.setList('tests', self.getTestNames())

    def gotResults_onetest(self, words, fullResults):
        """do one of the unittests"""
        test = words[-1]
        self.doAll = 0
        print("---------------------------------do unimacro unittest for: %s"% test)
        self.doUnitTests([test])

    def gotResults_alltests(self, words, fullResults):
        """do all the unittests"""
        print("---------------------------------do all unimacro unittests")
        tests = list(self.allTests.keys())
        tests.sort()
        self.doAll = 1
        self.doUnitTests(tests)

    def getTestNames(self):
        """get the test names, with side effect builds a dict of all the tests

        the result (the testNames) can then be filled in the list {tests}
        self.allTests (dict) contains the names: files entries
        """
        self.testFolder = os.path.join(status.getUnimacroDirectory(), "unimacro_test")
        testFiles = glob.glob(os.path.join(self.testFolder, "*test.py"))
##        print 'testFiles: %s'% testFiles
        testNames = list(map(self.extractTestName, testFiles))
        print('testNames: %s'% testNames)
        self.allTests = dict(list(zip(testNames, testFiles)))
        return testNames

    def extractTestName(self, fullPath):
        """remove path, .py and "Test" from filename to get pronouncable test"""
        name = os.path.basename(fullPath).lower()
        name = utilsqh.removeFromEnd(name, ".py", ignoreCase=1)
        name = utilsqh.removeFromEnd(name, "test", ignoreCase=1)
        return name

    def showInifile(self):
        commandExplanation = "Names of tests: \n%s\n---------------"% '\n'.join(self.testNames)
        super(UnittestGrammar, self).showInifile(commandExplanation=commandExplanation)

    def doUnitTests(self, tests):
        # self.checkSysPath(self.testFolder)  # append unimacro_test if needed
        suiteAll = None
        self.activeTests = []

        for test in tests:
            print('do test %s (%s)'% (test, self.allTests[test]))
            suite = self.addUnitTest(test, self.allTests[test])
            if suite:
                self.activeTests.append(test)
                if suiteAll:
                    suiteAll.addTest(suite)
                else:
                    suiteAll = suite
        if suiteAll:
            natlink.setMicState('off')
            result = unittest.TextTestRunner().run(suiteAll)
            print('after testing------------------------------------')
            self.dumpResult(result, filesToSkip=self.filesToSkipInResult)            
            natlink.setMicState('on')

        else:
            print("nothing valid to test")

    def addUnitTest(self, test, fileName):
        """do one of the unittests"""
##        actions.Message("starting test %s"% fileName)

        test_path = os.path.join(sys.prefix, 'Lib', 'site-packages', 'unimacro', 'unimacro_test')
        if not os.path.isdir(test_path):
            raise OSError(f'cannot add unittest for Unimacro, test_path invalid: "{test_path}"')
        if not test_path in sys.path:
            sys.path.append(test_path)
        modName = os.path.basename(fileName)
        modName = utilsqh.removeFromEnd(modName, ".py", ignoreCase=1)
    
        testMod = __import__(modName)
        testClassName = modName
        try:
            testClass = getattr(testMod, testClassName)
        except AttributeError:
            print('****cannot find test class in test module, skipping test: %s'% testClassName)
            return None
        suite = unittest.makeSuite(testClass, 'test')
        return suite

    def checkSysPath(self, folder):
        """add to sys.path if not present yet"""
        if folder in sys.path:
            return
        print('adding to path: %s'% folder)
        sys.path.append(folder)

    def dumpResult(self, testResult, filesToSkip=None):
        """dump into txt file "testresult" in unimacro_test folder

        slightly different version in voice coder (test_defs.py)
        """
        alert = len(self.activeTests) > 0
        testResultFile = os.path.join(self.testFolder, "testresult.txt")
        f = open(testResultFile, 'w')
        if testResult.wasSuccessful():
            if alert:
                mes = "all succesful: %s"% self.activeTests
                actions.Message(mes, "Unimacro Tests succesful")
            else:
                mes = "test passed: %s"% self.activeTests[0]
                self.DisplayMessage(mes)
            f.write(mes+'\n'*2)
            f.close()
            return
        f.write("results of tests: %s\n"% self.activeTests)
        f.write('\n--------------- errors -----------------\n')
        for case, tb in testResult.errors:
            f.write('\n---------- %s --------\n'% case)
            cleanTb = utilsqh.cleanTraceback(tb, filesToSkip)
            f.write(cleanTb)
        f.write('\n--------------- failures -----------------\n')
        for case, tb in testResult.failures:
            f.write('\n---------- %s --------\n'% case)
            cleanTb = utilsqh.cleanTraceback(tb, filesToSkip)
            f.write(cleanTb)

        f.close()
        mes = "there were errors with testing unimacro: %s\n\nPlease consult the outputfile: %s"% \
              (self.activeTests, testResultFile)
        actions.Message(mes, "Unimacro Tests failure", alert=alert)
        self.openFileDefault(testResultFile, mode="edit", name='test result')

    
        

    
# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
thisGrammar = UnittestGrammar()
if thisGrammar.gramSpec:
    thisGrammar.initialize()
else:
    thisGrammar = None

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None





# This was the old "general" intro for all unimacro grammar files:
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
#  _general.py
#
# written by Quintijn Hoogenboom (QH softwaretraining & advies),
#
#
#pylint:disable=C0302, R0904, C0209, C0321, R0912, R0914, R0915, R0911
#pylint:disable=E1101 

"""do a set of general commands

"""
import re
import os
import sys
import time
import pprint
from pathlib import Path

import win32gui

import natlink
from natlinkcore import natlinkstatus
from natlinkcore import nsformat 
from natlinkcore import natlinkutils as natut
from natlinkcore import natlinktimer
from unimacro import natlinkutilsbj as natbj
from unimacro import namelist # for name phrases

from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doKeystroke as keystroke
from dtactions.unimacro import unimacroactions as actions
from dtactions.unimacro import unimacroutils

# taskswitching moved to _tasks.py (july 2006)

Counts = list(range(1,20)) + list(range(20,51,5))

# for taskswitch:
Handles = {}
#systray:
systrayHndle = 0

status = natlinkstatus.NatlinkStatus()
language = status.language
FORMATS = {
    # for letters (do nothing):
    'no spacing': (unimacroutils.wf_NoSpaceFollowingThisWord | unimacroutils.wf_NoSpacePreceedingThisWord |
                   unimacroutils.wf_TurnOffSpacingBetweenWords |
                      unimacroutils.wf_DoNotApplyFormattingToThisWord
          ),
    # normal words:
    'normal words': ( unimacroutils.wf_RestoreNormalCapitalization |
            unimacroutils.wf_RestoreNormalSpacing
          ),
    # extra space(do one space):
    'extra space':  ( unimacroutils.wf_RestoreNormalCapitalization |
            unimacroutils.wf_RestoreNormalSpacing |
            unimacroutils.wf_AddAnExtraSpaceFollowingThisWord
          ), 
    }

user = status.user
unimacro_user_dir = status.getUnimacroUserDirectory()
wordsFolder = f'{unimacro_user_dir}\\{language}_words'
if not Path(wordsFolder).is_dir():
    Path(wordsFolder).mkdir()
files = [os.path.splitext(f)[0] for f in os.listdir(wordsFolder)]
## print '_general, files in wordsFolder %s: %s'% (wordsFolder, files)

if language == 'enx':
    nameList = {'Q. H.': 'QH',
                 'R. A.': 'RA',
                'underscore': '',
                 }
elif language == 'nld':
    nameList = {'QH': 'QH',
                 'er aa': 'RA',
                'underscore': '',
                 }
else:
    nameList = {}
        

switchDirection = {
      "{Up}":      "{Down}",
      "{Down}":    "{Up}",
      "{Left}":    "{Right}",
      "{Right}":   "{Left}"}

modes = ['spell', 'command', 'numbers', 'normal', 'dictation', 'dictate']
normalSet = ['test', 'reload', 'info', 'undo', 'redo', 'namephrase', 
             'comment', 'documentation', 'modes', 'variable', 'search',
             'highlight',         # for Shane, enable, because Vocola did not fix _anything yet
             'browsewith', 'hyphenatephrase', 'pastepart',
             'password', 'choose']
#normalSet = ['hyphenatephrase']  # skip 'openuser'

commandSet = normalSet[:] + ['dictate']
thisGrammar = None

ancestor=natbj.IniGrammar
class ThisGrammar(ancestor):
    # pylint: disable=C0116, W0613, W0201

    iniIgnoreGrammarLists = ['modes','count', 'namelist', 'character', 'punctuation']

    language = unimacroutils.getLanguage()        

    try:
        number_rules = natbj.numberGrammar[language] #  including millions
    except KeyError:
        number_rules = natbj.numberGrammar['enx']

    name = "general"
    gramSpec = ["""
<before> = Command | Here;
<dgnletters> imported;
<dgndictation> imported;
# <dgnwords> imported;
<documentation> exported = Make documentation;
<test> exported = test micstate;
<choose> exported = choose {n1-10};
<reload> exported = reload Natlink;
<info> exported = give (user|prog|window|unimacro|path|timer) (info|information) ;
<undo> exported = Undo [That] [{count} [times]];
<redo> exported = Redo [That] [{count} [times]];
<namephrase> exported = Make That [Name] phrase;
<pastepart> exported = [<before>] Paste part {n1-10};
<pasteallparts> exported = Paste all parts;
<hyphenatephrase> exported = Hyphenate (phrase| last word | last ({n2-5}) words);
<comment> exported = Comment {namelist};
<variable> exported = {formatvariable} ((Back [{n1-5}]) | <dgndictation>);
<modes> exported = {modes} mode;
<highlight> exported = highlight <dgndictation>;         # for Shane,
<search> exported = search ('go back'|new|
                            ((for|before|after|extend|insert)([{searchwords}] (<dgndictation>|<characterpunctuation>+)))|
                            ((forward|back|up|down|reverse) [{count} [times]]));
<characterpunctuation> = space|capital|{character}|{punctuation};
<browsewith> exported = ('browse with') {browsers};
<openuser> exported = 'open user' {users};
<password> exported = 'password' <dgndictation>;
<presscode> exported = (press letter) <dgnletters>;

    """]
# A\determinerNormalTestThisNowI\pronounTest.\period\full stop
#NormalTestNormalI\pronounTestNormallyTestNormalAttestNormal.\period\periodSigns
#HelloTesting
    
    def initialize(self):
        if self.language:
            self.load(self.gramSpec)
            self.switchOnOrOff(activateSet=normalSet) # initialises lists from inifile, and switches on
                             # if all goes well (and variable onOrOff == 1)
            # search commands:
            self.setCharactersList('character')
            self.setPunctuationList('punctuation')
            self.specialSearchWords = self.Lists['searchwords'] or [] # like function, class of (inifile) section
            # print('specialSearchWords: %s'% self.specialSearchWords)
            self.setNumbersList('count', Counts)
            self.setList('modes', modes)
            self.gotPassword = 0
            self.passwordEnding = None
            self.gotPresscode = 0
            # print "%s, activateSet: %s"% (self.name, normalSet)
            # self.deactivateAll()  # why is this necessary? The activateAll in switchOn is definitly now Ok...
            self.title = 'Unimacro grammar "'+__name__+'" (language: '+self.language+')'
        else:
            print("no valid language in grammar "+__name__+" grammar not initialized")

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile() # refills grammar lists and instance variables
                                # if something changed.
        self.gotPassword = 0
        self.progInfo = unimacroutils.getProgInfo(moduleInfo)
        
    def gotResults_wrongrule(self,words,fullResults):
        natut.playString("%s\n"% fullResults)

    def gotResultsInit(self,words,fullResults):
        self.fullText = ' '.join(words)
        self.progName = unimacroutils.getProgName()

        # variable formatting
        self.gotVariable = ""

        # for Shane
        self.search = self.dictate = self.highlight = 0
        self.text = ''
        self.minimalapp=None
        # self.specialSearchWords = None

        if words[0] in ['Hier', 'Here']:
            print('Here from _general...')
            natut.buttonClick()
            unimacroutils.Wait()
        self.progInfo = unimacroutils.getProgInfo()
    def gotResults_password(self,words,fullResults):
        """interpret password as dictate
        Cap dictation words
        if number precedes @ 
        
        """
        self.gotPassword = 1

    def gotResults_pastepart(self,words,fullResults):
        """paste part of clipboard, parts are separated by ";"
        """
        n = self.getNumberFromSpoken(words[-1])
        t = unimacroutils.getClipboard()
        print('(%s) %s'% (type(t), t))
        T = self.partsSplitSpecial(t)
        if n <= len(T):
            keystroke(T[n-1])
            # action("SCLIP %s"% T[n-1])
        else:
            print('_general, pastepart: length of list only: %s (t: %s)'% (len(T), t))

    def gotResults_pasteallparts(self,words,fullResults):
        """paste parts, separated by ";" with a hard set keystroke in between
        
        Used for pasting multiple addresses in Thunderbird address book
        """
        # n = self.getNumberFromSpoken(words[-1])
        t = unimacroutils.getClipboard()
        print('(%s) %s'% (type(t), t))
        T = self.partsSplitSpecial(t)
        print('put item by item %s words'% len(T))
        for t in T:
            keystroke(t)
            keystroke("{enter}")
            action("VW")

    def partsSplitSpecial(self, t):
        """normally split by ;
        in special cases, eg for coordinates the split can be different
        """
        reCoords = re.compile(r"(.*?) N([0-9\.]+) +E([0-9\.]+)$")
        # Title of place XX N51.5888  E5.901
        m = reCoords.match(t)
        if m:
            parts = m.groups()
            print('coordinates: %s, length: %s'% (repr(parts), len(parts)))
        elif t.find(";") >= 0:
            parts = [t.strip() for t in t.split(";")]
            print('splitted string: %s, length: %s'% (repr(parts), len(parts)))
        else:
            parts = [t]
            print('cannot split text: %s\nreturn list of length 1: %s'% (t, parts))
        return parts
        
    def gotResults_before(self,words,fullResults):
        if self.hasCommon(words, 'here'):
            natut.buttonClick('left', 1)

    def gotResults_highlight(self,words,fullResults):
        # for Shane
        self.highlight = 1

    def gotResults_search(self,words,fullResults):
        self.search = 1
        if words[0] in self.specialSearchWords:
            self.specialSearchWord = self.getFromInifile(words[0], 'searchwords')
            print('do search with special search word: %s (%s)'% (self.specialSearchWord, words[0]))
            words.pop(0)

        counts = self.getNumbersFromSpoken(words, Counts)
        if counts:
            self.count = counts[0]
        else:
            self.count = None
        if self.hasCommon(words, ['new', 'nieuw']):
            self.search = 'new'
        elif self.hasCommon(words, ['back', 'up', 'omhoog', 'terug', 'achteruit']):
            self.search = 'back'
        elif self.hasCommon(words, ['reverse']):
            self.search = 'reverse'
        elif self.hasCommon(words, ['forward', 'down', 'vooruit', 'omlaag', 'verder']):
            self.search = 'forward'
        elif self.hasCommon(words, ['go back', 'ga terug']):
            self.search = 'go back'
        elif self.hasCommon(words, ['for', 'naar']):
            self.search = 'for'
        elif self.hasCommon(words, ['before', 'voor']):
            self.search = 'before'
        elif self.hasCommon(words, ['after', 'na']):
            self.search = 'after'
        elif self.hasCommon(words, ['insert', 'invoegen']):
            self.search = 'insert'
        elif self.hasCommon(words, ['extend', 'uitbreiden']):
            self.search = 'extend'

        # provisions for extra special keywords (in front of spelled characters)            
        self.specialSearchWord = self.hasCommon(words, self.specialSearchWords)


    def gotResults_dictate(self,words,fullResults):
        self.dictate = 1

    def gotResults_dgnletters(self,words,fullResults):
        self.text = ''.join(map(unimacroutils.stripSpokenForm, words))
        if self.search:
            # catch some common misrecognitions:
            if self.text == '4':
                print(f'caught dgnletters {self.text}, switch to forward search')
                self.text = ''
                self.search = 2 # forward search
            elif self.text in ['43', '403']:
                print(f'caught dgnletters {self.text}, switch to forward search 3')
                self.text = ''
                self.count = 3
                self.search = 2 # forward search
            return
        if self.gotPresscode:
            print(f'gotPresscode: {words} -> {self.text}')
            self.do_pressfirst(self.text)
            return
        
        
    def gotResults_characterpunctuation(self,words,fullResults):
        capNext = 0
        for w in words:
            if self.hasCommon(w, "capital"):
                print('got word (synonym of) "capital": %s'% w)
                capNext = 1
                continue
            if self.hasCommon(w, "space"):
                self.text += " "
                capNext = 0
                continue
            char = self.getCharacterFromSpoken(w)
            if char:
                if capNext:
                    self.text += char.upper()
                    capNext = 0
                else:
                    self.text += char
            else:
                capNext = 0
                char = self.getPunctuationFromSpoken(w)
                if char:
                    self.text += char
                else:
                    print('general: character or punctuation not found for spoken form: %s'% w)
        
    # def gotResults_dgnwords(self,words,fullResults):
    #     #self.text = ' '.join(map(unimacroutils.stripSpokenForm, words))
    #     # try with the improved nsformat function
    #     print(f'got dgnwords: {words}')

    def gotResults_dgndictation(self,words,fullResults):
        #self.text = ' '.join(map(unimacroutils.stripSpokenForm, words))
        # try with the improved nsformat function 
        if self.gotPassword:
            print('gotPassword, analyse password: %s'% words)
            text = nsformat.formatPassword(words)
            keystroke(text)
            self.gotPassword = 0
            return
        if self.gotVariable:
            print('do variable trick %s on %s'% (self.gotVariable, words))
            vartrick = self.gotVariable
            funcName = 'format_%s'% vartrick
            # print 'funcName: %s'% funcName
            try:
                func = getattr(self, funcName)
                # print 'func: %s'% func
            except AttributeError:
                print('no formatfunction for variable trick: %s'% vartrick)
                return

            result = func(words)
#
            keystroke(" " + result)
            return
        if self.gotPresscode:
            self.text = ' '.join(map(unimacroutils.stripSpokenForm, words))
            print(f'got dgndictation: {words} -> {self.text}')
            self.do_pressfirst(self.text)
            return
        #very well for like this
        if self.search and self.text in ['on', 'verder']: # 
            self.search = 2
        elif self.search and self.text in ['new', 'nieuw']:
            self.search = 3
        elif self.search and self.text in ['terug', 'back']:
            self.search = 4
        print('dgndictation: %s'% self.text)

    def format_camel(self, words):
        """format camel case, rule variable
        var like this -> varLikeThis
        """
        if not words: return ""   #
        newWords = [w.capitalize() for w in words]
        newWords[0] = newWords[0].lower()
        return ''.join(newWords)

    def format_studly(self, words):
        """format studly case, rule variable
        var like this -> VarLikeThis
        """
        if not words: return ""   #
        newWords = [w.capitalize() for w in words]
        return ''.join(newWords)

    def format_dotword(self, words):
        """format dotword, rule variable
        var like this -> var.like.this
        """
        if not words: return ""   #
        return '.'.join(words)

    def format_jive(self, words):
        """format jive case, rule variable
        var like this -> var-like-this
        """
        if not words: return ""   #
        return '-'.join(words)

    def format_score(self, words):
        """format score, with underscores, rule variable
        var like this -> var_like_this
        """
        if not words: return ""   #
        return '_'.join(words)

# try for like this

#
    def gotResults_browsewith(self,words,fullResults):
        """show page in another browser"""
        progInfo = unimacroutils.getProgInfo()
        prog = progInfo.prog
        # Iam2x = prog == '2xexplorer'
        # IamExplorer = prog == 'explorer'
        browser = prog in ['iexplore', 'firefox','opera', 'netscp', 'chrome']
        if not browser:
            self.DisplayMessage ('command only for browsers')
            return
        print('words:', words)
        unimacroutils.saveClipboard()
        action('<<addressfield>>; {extend}{shift+exthome}{ctrl+c};<<addressfieldclose>>')
        askedBrowser = self.getFromInifile(words, 'browsers')
        if askedBrowser == prog:
            self.DisplayMessage('command only for another browser')
            return
        print('try to bring up browser: |%s|'% askedBrowser)
        action('RW')
        action('AppBringUp "%s"'% askedBrowser)
        action('WTC')
        action('<<addressfield>>; {ctrl+v}{enter}')
        
        unimacroutils.restoreClipboard()
 
    def gotResults_documentation(self,words,fullResults):
        print("obsolete")
#         oldPath = os.getcwd()
#         uniGrammars = self.ini.getList('documentation', 'unimacro grammars')
#         uniModules = self.ini.getList('documentation', 'unimacro modules')
#         otherGrammars = self.ini.getList('documentation', 'other grammars')
#         otherModules = self.ini.getList('documentation', 'other modules')
#         base = unimacroutils.getUnimacroUserDirectory()
#         docPath = os.path.join(base, 'doc')
#         pickleFile = os.path.join(docPath, '@unimacro.pickle')
#         try:
#             psock = open(pickleFile, 'r')
#             memory = pickle.load(psock)
#             psock.close()
#             print('--------------------memory from pickle: %s'% pickleFile)
#         except:
#             memory = {}
#             print('--------------------no or invalid pickle file: %s'% pickleFile)
#             
#         utilsqh.createFolderIfNotExistent(docPath)
#         os.chdir(docPath)
#         self.DisplayMessage('writing documentation to: %s'% docPath)
#         pydoc.writedocs(base)
#         self.DisplayMessage('checking unimacro grammars, modules and other grammars, modules')
#         allUnimacroGrammars = list(natlinkmain.loadedFiles.keys())
#         if 'unimacro grammars' not in memory:
#             memory['unimacro grammars'] = {}
#         mem = memory['unimacro grammars']
#         for m in uniGrammars:
#             if m in allUnimacroGrammars:
#                 mem[m] = sys.modules[m].__doc__
#             else:
#                 if not m in mem:
#                     mem[m] = ''
# 
#         if 'unimacro modules' not in memory:
#             memory['unimacro modules'] = {}
#         mem = memory['unimacro modules']
#         for m in uniModules:
#             if m in sys.modules:
#                 mem[m] = sys.modules[m].__doc__
#             else:
#                 try:
#                     M = __import__(m)
#                 except ImportError:
#                     print('cannot import module: %s'% m)
#                     continue
#                 mem[m] = M.__doc__
#                 mem[m] = M.__doc__
#                 del M
# 
#         print('writing to pickle file: %s'% pickleFile)
#         psock = open(pickleFile, 'w')
#         pickle.dump(memory, psock)
#         psock.close()
#         L = []
#         htmlFiles = list(filter(isHtmlFile, os.listdir(docPath)))
#         
#         
#         categories = self.ini.get('documentation')
#         if not categories:
#             self.DisplayMessage('please fill in documentation categories')
# 
#         for c in categories:
#             if not c in memory:
#                 continue
#             L.append("<H1>%s</H1>"% c)
#             mem = memory[c]
#             for m in mem:
#                 file = m+'.html'
#                 if os.path.isfile(os.path.join(docPath, m+'.html')):
#                     link = "<a href=%s.html>%s</a>"% (m, m)
#                     htmlFiles.remove(file)
#                 else:
#                     link = "???%s"% m
#                 if mem[m] == None:
#                     text = 'no doc string for this module'
#                 elif mem[m] == '':
#                     text = 'module could not be loaded, possibly start program and do "Make documentation" again'
#                 else:
#                     text = mem[m]
# 
#                 if text.find('\n\n'):
#                     T = text.split('\n\n')
#                     text = T[0]
#                 L.append("<p>%s: %s</p>"% (link, text))
#         if htmlFiles:
#             M = []
#             L.append("<H1>%s</H1>"% "other files")
#             for f in htmlFiles:
#                 if f == 'index.html':
#                     continue
#                 name = f.split('.')[0]
#                 link = "<a href=%s>%s</a>"% (f, name)
#                 M.append(link)
#             L.append("<p>%s</p>"% ', '.join(M))
#         HTMLpage = '''<!doctype html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
# <html><head><title>Natlink grammars and modules documentations</title>
# <style type="text/css"><!--
# TT { font-family: lucidatypewriter, lucida console, courier }
# --></style></head><body bgcolor="#f0f0f8">
# %s
# </body></html>''' % '\n'.join(L)
#         fsock = open(os.path.join(docPath, 'index.html'), 'w')
#         fsock.write(HTMLpage)
#         fsock.close()
#                     
#         os.chdir(oldPath)
#         
#         self.DisplayMessage('okay')
        

    def gotResults_stopwatch(self,words,fullResults):
        """ stopwatch"""
        if self.hasCommon(words, 'start'):
            self.startTime = time.time()
        else:
            t = time.time()
            elapsed = t - self.startTime
            action(f'MSG {elapsed:.2f} seconds')
            self.startTime = t

    def gotResults_choose(self,words,fullResults):
        """choose alternative, via actions
                
        """
        n = self.getNumberFromSpoken(words)   # return int
        print(f'got choose: {words} -> {n}')
        action(f'<<choose {n}>>')
        
        
    def gotResults_test(self,words,fullResults):

        # micstate = natlink.getMicState()
        for ms in ('off', 'on'):
            print("switching %s mic"% ms)
            natlink.setMicState(ms)
            time.sleep(1)
            newMs = natlink.getMicState()
            time.sleep(1)
            if ms == newMs:
                time.sleep(0.5)
                continue
            print("conflicting mic states, now: %s, expected: %s"% (newMs, ms))

    def getPrevNext(self, n=1):
        """return character to the left and to the right of the cursor
        assume no selection active.
        normally return cursor in same position
        
        This one gives timing problems in Frescobaldi (lilypond edit program), the ctrl+c takes about one second.
        now try in other applications
        """
        playString = natut.playString
        prog = unimacroutils.getProgInfo()[0]
        t0 = time.time()
        unimacroutils.clearClipboard()
        t1 = time.time()
        playString("{left %s}"% n)
        t2 = time.time()
        playString("{shift+right %s}"% (n*2,))
        t3 = time.time()
        playString("{ctrl+c}")
        t4 = time.time()
        playString("{left %s}"% n)
        t5 = time.time()
        result = unimacroutils.getClipboard()
        t6 = time.time()
        print('timing getPrevNext program: %s\nclear clipboard: %.4f, left: %.4f, shiftright2: %.4f, copy: %.4f, left: %.4f, getcl: %.4f'% (
            prog, t1-t0, t2-t1, t3-t2, t4-t3, t5-t4, t6-t5))
        if len(result) == 2:
            return result[0], result[1]
        if result == '\n':
            print('getPrevNext, assume at end of file...')
            # assume at end of file, could also be begin of file, but too rare too handle
            playString("{right}")
            return result, result
        print('getPrevNext, len not 2: %s, (%s)'% (len(result), repr(result)))
        return "", result

##        
    def gotResults_reload(self,words,fullResults):
        print("reloading natlink....")
        unimacroutils.switchToWindowWithTitle("Messages from Python Macros")
        unimacroutils.Wait()
        natlink.setMicState("off")
        unimacroutils.Wait()
        print("do it yourself...")
    
   # deze regel print de naam van de huidige module in het debug-venster
    def gotResults_info(self,words,fullResults):
        """display in a message box information about the window, user or unimacro


        """
        T = []
        extra = []
        if self.hasCommon(words,'window') or self.hasCommon(words,'prog'):
            p = unimacroutils.getProgInfo()
            hndle = p.hndle
            T.append('---from unimacroutils.getProgInfo:')
            # (progpath, prog, title, toporchild, classname, hndle)
            if self.hasCommon(words, 'prog'):
                T.append(f'progInfo = {p}')
            T.append(f'  .progpath: "{p.progpath}"\n')
            T.append(f'  .prog:\t{p.prog}')
            T.append(f'  .title: "{p.title}"\n')
            T.append(f'  .toporchild\t{p.toporchild}')
            T.append(f'  .classname\t{p.classname}')
            T.append(f'  .hndle:\t{hndle}')
            # # for special behaviour:
            # childClass = "#32770"
            # overruleIsTop = self.getTopOrChild(self.progInfo, childClass=childClass)
            # 
            # if p.toporchild != overruleIsTop:
            #     T.append('')
            #     if overruleIsTop:
            #         T.append("**** treat as TOP window although it is a child window")
            #     else:
            #         T.append(f'**** treat as CHILD window although it is a top window (classname: {classname})')


        elif self.hasCommon(words,'user'):
            # status (natlinkstatus.NatlinkStatus()) is global variable
            T.append('user:\t\t%s'% status.user)
            T.append('userLanguage:\t%s'% status.language)
            # T.append('see messages window for trainuser info')
            extra = []
            
        elif self.hasCommon(words,'unimacro'):
            # status (natlinkstatus.NatlinkStatus()) is global variable
            version = status.getDNSVersion()
            T.append('DNSVersion:\t\t%s'% version)
            wVersion = status.getWindowsVersion()
            T.append('WindowsVersion:\t\t%s'% wVersion)
            T.append('UnimacroDirectory:\t%s'% status.getUnimacroDirectory())
            T.append('UnimacroUserDirectory:\t%s'% status.getUnimacroUserDirectory())
            T.append('UnimacroGrammarsDirectory:\t%s'% status.getUnimacroGrammarsDirectory())
        elif self.hasCommon(words,'path'):
            T.append('the python path:')
            T.append(pprint.pformat(sys.path))
        elif self.hasCommon(words, "timer"):
            timer_info = natlinktimer.getNatlinktimerStatus()
            if timer_info is None:
                T.append('natlinktimer is not active')
            elif timer_info == 1:
                T.append('there is 1 timer active')
            else:
                T.append(f'there are {timer_info} timers active')
        elif self.hasCommon(words, "class"):
            T.append()
        else:
            T.append('no valid keyword found')

        s = '\n'.join(T)
            
        actions.Message(s)
        print(s)
        print()
        for e in extra:
            print(e)

    def gotResults_variable(self,words,fullResults):
        vartrick = self.getFromInifile(words[0], 'formatvariable', '')
        print('vartrick: %s'% vartrick)
        if vartrick:
            self.gotVariable = vartrick
        else:
            print('no vartrick found, return')
        # 
        c = self.getNumberFromSpoken(words[-1])
        
        if not c:
            print('vartrick %s wait for dgndictation'% vartrick)
            return
        keystroke('{Shift+Ctrl+Left %s}' % c)
        keystroke('{ctrl+x}')
        print('here comes the copy paste trick %s words'% c)
        unimacroutils.Wait(0.5)
        t = natlink.getClipboard()
        tList = t.split()
        print('tList: %s'% tList)
        unimacroutils.Wait(0.5)
        funcName = 'format_%s'% vartrick
        # print 'funcName: %s'% funcNameyour 
        try:
            func = getattr(self, funcName)
            # print 'func: %s'% func
        except AttributeError:
            print('no formatfunction for variable trick: %s'% vartrick)
            return

        result = func(tList)
#
        keystroke(result)
        return


    def capit(self, s):
        """ capitalise, but leave upper case characters as they are

        (the builtin capitalise makes first letter upper case and the following
         letters lowercase)
        """
        return s[0].upper() + s[1:]

    def gotResults_undo(self,words,fullResults):
        counts = self.getNumbersFromSpoken(words)
        if counts:
            count = counts[0]
        else:
            count = 1
        #print 'count: %s'% count
        for _ in range(count):
            action('<<undo>>')

    def gotResults_redo(self,words,fullResults):
        counts = self.getNumbersFromSpoken(words)
        if counts:
            count = counts[0]
        else:
            count = 1
        #print 'count: %s'% count
        for _ in range(count):
            action('<<redo>>')

    def gotResults_comment(self,words,fullResults):
        name = nameList[words[-1]]
        if name:
            ts = time.strftime("%d%m%Y", time.localtime(time.time()))
        else:
            ts = time.strftime("%d%m%y_", time.localtime(time.time()))

        progInfo = unimacroutils.getProgInfo()
        if unimacroutils.matchModule('pythonwin', progInfo=progInfo):
            com = "#" + name + ts
        elif unimacroutils.matchModule('textpad', 'html', progInfo=progInfo):
            com = "<!--" + name + ts + "-->"
        elif unimacroutils.matchModule('textpad', '.c', progInfo=progInfo):
            com = "$$$$" + name + ts + "$$$$"
        elif unimacroutils.matchModule('textpad', '.py', progInfo=progInfo):
            com = "#" + name + ts
        else:
            com = name + ts
        keystroke(com+"\n")
            
    def gotResults_modes(self,words,fullResults):
        """enable different modes

        When going to spell mode or command mode the special
        set which makes searching and dictating a single word
        or a few words possible are enabled.

        """
        mode = self.hasCommon(words, modes)
        if not mode:
            print('modes, invalid mode: %s'% words)
            return
        if mode in ['normal', 'normale']:
            M = 0
        elif mode in ['dictation', 'dicteer', 'dictate']:
            M = 1
        elif mode in ['command', 'commando']:
            M = 2
        elif mode in ['numbers', 'nummer']:
            M = 3
        elif mode in ['spell', 'spel']:
            M = 4
        else:
            print('no valid mode: %s'% mode)
            return

        self.setMode(M)

        if M > 1:
            self.DisplayMessage('<_general: setting command set>')
            self.activateSet(commandSet)
        else:
            self.DisplayMessage('<_general: setting normal set>')
            self.activateSet(normalSet)
            
        

    def gotResults_namephrase(self,words,fullResults):
        # list of words that can be combined in a double christian name
        #  eg Jan Jaap or Jan-Marie 
        # voornamenList = ['Jan', 'Jaap', 'Peter', 'Louise', 'Anne']
        modInfo = natlink.getCurrentModule()
        action("CLIPSAVE")
        keystroke("{Ctrl+c}")
        time.sleep(0.1)
        # do contents of clipboard:
        t = natlink.getClipboard().strip()
        if not t:
            modInfo = natlink.getCurrentModule()
            if unimacroutils.matchModule('natspeak', 'spell', modInfo):
                keystroke("{ExtHome}{Shift+ExtEnd}{Ctrl+x}")
                unimacroutils.Wait(0.5)
                t = natlink.getClipboard().strip()
                if not t:
                    action("CLIPRESTORE")
                    return
            else:
                if self.language == 'nld':
                    com = "selecteer dat"
                else:
                    com  = "select that"
                action("HW %s"%com)
                unimacroutils.Wait(0.5)
                keystroke("{Ctrl+c}")
                time.sleep(0.1)
                t = natlink.getClipboard().strip()
                if not t:                    
                    self.DisplayMessage("select a text first")
                    action("CLIPRESTORE")
                    return
        if self.hasCommon(words, ['naam', 'Name']):
            result = namelist.namelistUnimacro(t, ini=self.ini)
            print('result of namelistUnimacro function: %s'% result)
            r = '' # in case result is empty
            for r in result:
                print(f'adding part: {r}')
                unimacroutils.addWordIfNecessary(t)
            keystroke(r)
        else: # zonder naam in words, a normal phrase:
            print('adding phrase %s'% t)
            unimacroutils.addWordIfNecessary(t)
            keystroke(t)
        action("CLIPRESTORE")

            
    def gotResults_hyphenatephrase(self,words,fullResults):
        # selection or last utterance is spelled out with all caps and hyphens
        # Quintijn, August 15, 2009

        # save clipboard, release after the action:                                 
        action("CLIPSAVE")
        # hasCommon function for possibility of translations/synonyms without altering the code:
        if self.hasCommon(words[-1], "phrase"):
            keystroke("{Ctrl+c}")
            # do contents of clipboard:
            t = natlink.getClipboard().strip()
            if not t:
                if self.language == 'nld':
                    com = "selecteer dat"
                else:
                    com  = "select that"
                action("HW %s"%com)
                unimacroutils.Wait(0.5)
                keystroke("{Ctrl+c}")
                t = natlink.getClipboard().strip()
            if not t:                    
                self.DisplayMessage("select a text first")
                action("CLIPRESTORE")
                return
        elif self.hasCommon(words, ('word', 'words')):
            counts = self.getNumbersFromSpoken(words)
            if counts:
                count = counts[0]
            else:
                count = 1
            keystroke("{shift+ctrl+left %s}"% count)
            keystroke("{Ctrl+c}")
            t = natlink.getClipboard().strip()
            if not t:                    
                self.DisplayMessage("could not select a valid text")
                action("CLIPRESTORE")
                return
        else:
            self.DisplayMessage("unexpected last word in command phrase: %s"% words[-1])
            action("CLIPRESTORE")
            return
            

        # first paste back the selected text, and add a space if needed:
        L = []
        # for each word in utterance join uppercased characters of word with a '-'
        tWords = t.split()
        for word in tWords:
            L.append('-'.join([t.upper() for t in word]))
            L.append(' ')
        keystroke(''.join(L))
        action("CLIPRESTORE")
        # 
    def gotResults_openuser(self,words,fullResults):
        User = self.getFromInifile(words[-1], 'users')
        print('user: %s'% User)
        try:
            natlink.openUser(User)
        except natlink.UnknownName:
            print(f'cannot open user "{User}", unknown name')
            
    def gotResults(self,words,fullResults):
        if self.highlight:
            # for Shane
            asterisksSpacing = 1   # to be perfected later as option of this grammar
            if asterisksSpacing:
                if self.text.find('*'):
                    self.text = self.text.replace('*', ' * ')
            if self.text:
                action("<<startsearch>>")
                keystroke(self.text)
                action("<<searchgo>>")    
            else:
                print('no text to highlight')
            return
            #keystroke("{ctrl+f}")
            #t = self.text
            #t = t.replace(' . ', '.')
            #t = t.replace('( ', '(')
            #t = t.replace(' )', ')')
            ##print 'execute highlight with "%s"'% highlightText
            #keystroke(t)
            #keystroke("{enter}")
            #return

        if self.search:
            progInfo = unimacroutils.getProgInfo()

            # make provisions for searchwords (function (def), class (class) etc)
            if self.specialSearchWord:
                self.text = self.ini.get('searchwords', self.specialSearchWord) + self.text
            if self.count:
                count = int(self.count)
            else:
                count = 1
            if self.search == 'forward':
                # forward
                self.direc = 'down'
                self.searchOn(count, progInfo=progInfo)
                return
            if self.search == 'new':
                # new, just start the search dialog:
                self.searchMarkSpot(progInfo=progInfo)
                action('<<startsearch>>', progInfo=progInfo)
                return
            if self.search == 'back':
                # back:
                self.direc = 'up'
                self.searchOn(count, progInfo=progInfo)
                return
            if self.search ==  'go back':
            # go back, return to origin
                print("search go back")
                self.searchGoBack(progInfo=progInfo)
                return
            if self.search in ('for', 'before','after'):
                # new search with text
                self.direc = 'down'
                print('new leap to text: %s'% self.text)
                self.searchMarkSpot(progInfo=progInfo)
                res = self.searchForText(self.direc, self.text, progInfo=progInfo, beforeafter=self.search)
            elif self.search == 'extend':
                res = self.searchForText(self.direc, self.text, progInfo=progInfo, extend=1)
            elif self.search == 'insert':
                res = self.searchForText(self.direc, self.text, progInfo=progInfo, insert=1)
            else:
                print('invalid search code: %s'% self.search)
                self.DisplayMessage('search, invalid search code: %s'% self.search)
                return
            if res == -2:
            # search failed, did cancel mode
                return 
            unimacroutils.visibleWait()
            print('calling stop search')
            self.stopSearch(progInfo=progInfo)
            

    def searchOn(self, count, progInfo=None):
        """search up or down possibly more times"""
        if progInfo is None:
            progInfo = unimacroutils.getProgInfo()
        sectionList = actions.getSectionList(progInfo=progInfo)
        if self.direc == 'back':
            searchGoOn = actions.getMetaAction('searchgoback', sectionList=sectionList, progInfo=progInfo)
        else:
            searchGoOn = actions.getMetaAction('searchgoforward', sectionList=sectionList, progInfo=progInfo)
            
        for _ in range(count):
            if searchGoOn:
                res = action(searchGoOn)
            else:
                res = self.searchForText(self.direc, progInfo=progInfo)
            self.direc = self.getLastSearchDirection() # in case back search changed it!
            if res == -2:
                # search failed, did cancel mode
                return 
        unimacroutils.visibleWait()
        if not searchGoOn:
            self.stopSearch(progInfo)


    def GetGrammarModuleName(self):
        return __name__    

    def GetDictionaries(self):
        Dicts={
        }
        return Dicts

    def Message(self,t):
        tt = t + "  (command: " + self.fullText + ")"
        unimacroutils.Message(tt,self.title)
        
    def do_pressfirst(self, text):
        """first character "hard", rest normal
        """
        if text:
            action(f'SSK {text[0]}')
            if len(text) > 1:
                action('VW')
                keystroke(text[1:])

def isPythonFile(f):
    return f[-3:] == '.py'

def isHtmlFile(f):
    return f[-5:].lower() == '.html'


Classes = ('TkTopLevel')
def getIdleTitles():
    """get all titles of top windows with class name in tuple below

    This class name belongs, as far as I know, to the window explorer window

    """
    TitlesHandles = []

    ##print 'Classes:', Classes
##    Classes = None
    win32gui.EnumWindows(getIdleWindowsWithText, (TitlesHandles, Classes))
    return TitlesHandles

def getIdleWindowsWithText(hwnd, th):
    TH, classes = th
##    if wTitle.find('d:') == 0:
##        print 'class:', win32gui.GetClassName(hwnd)
    if win32gui.GetClassName(hwnd) in classes:
        wTitle = win32gui.GetWindowText(hwnd).strip().lower()
        TH.append((wTitle, hwnd))

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None
#
if __name__ == "__main__":
    # here code to interactive run this module
    thisGrammar = ThisGrammar(inifile_stem='_general')
    thisGrammar.startInifile()
    thisGrammar.initialize()
elif __name__.find('.') == -1:
    # this is caught when this module is imported by the loader (when Dragon/Natlink starts)
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()





"""Extracts the threadinfo of the current thread. Can see if a menu
or popup is active

hasMenuOpen() gives true if menu (File, Edit, ...) is open or a right click
context menu is open.
hasPopupOpen() give true if a right click (context) menu is open.

getWindowFlags gives (as a tuple) the flags that are set in the guithreadinfo.flags variable.


Used for grammar kaiser_dictation.py in conjunction with unimacro grammars
(qh.antenna.nl/unimacro). Quintijn Hoogenboom, december 2009
"""

from ctypes import *
import types
user32 = windll.user32
kernel32 = windll.kernel32
import time
class RECT(Structure):
 _fields_ = [
     ("left", c_ulong),
     ("top", c_ulong),
     ("right", c_ulong),
     ("bottom", c_ulong)
 ]

class GUITHREADINFO(Structure):
 _fields_ = [
     ("cbSize", c_ulong),
     ("flags", c_ulong),
     ("hwndActive", c_ulong),
     ("hwndFocus", c_ulong),
     ("hwndCapture", c_ulong),
     ("hwndMenuOwner", c_ulong),
     ("hwndMoveSize", c_ulong),
     ("hwndCaret", c_ulong),
     ("rcCaret", RECT)
 ]

def moveCursorInCurrentWindow(x, y):
 # Find the focussed window.
    guiThreadInfo = GUITHREADINFO(cbSize=sizeof(GUITHREADINFO))
    user32.GetGUIThreadInfo(0, byref(guiThreadInfo))
    focussedWindow = guiThreadInfo.hwndFocus

    # Find the screen position of the window.
    windowRect = RECT()
    user32.GetWindowRect(focussedWindow, byref(windowRect))

    # Finally, move the cursor relative to the window.
    user32.SetCursorPos(windowRect.left + x, windowRect.top + y)

def hasMenuOpen():
    guiThreadInfo = GUITHREADINFO(cbSize=sizeof(GUITHREADINFO))
    user32.GetGUIThreadInfo(0, byref(guiThreadInfo))
    flags = guiThreadInfo.flags
    menuOpen = 4
    return (flags & menuOpen) == menuOpen

def hasPopupOpen():
    guiThreadInfo = GUITHREADINFO(cbSize=sizeof(GUITHREADINFO))
    user32.GetGUIThreadInfo(0, byref(guiThreadInfo))
    flags = guiThreadInfo.flags
    popupOpen = 16
    return (flags & popupOpen == popupOpen)
    
def getWindowFlags():
    guiThreadInfo = GUITHREADINFO(cbSize=sizeof(GUITHREADINFO))
    user32.GetGUIThreadInfo(0, byref(guiThreadInfo))
    return flagsToTuple(guiThreadInfo.flags)
    
def flagsToTuple(flags):
    """convert a flags int into a set of flags
    """
    if flags == None:
        return ()
    elif flags == 0:
        return ()
    Flags = []
    if type(flags) in (int, int):
        if type(flags) == int: n = 32
        elif type(flags) == int: n = 64
        else:
            raise ValueError('type should be "int" or "long" here, not: %s'% type(flags))
        if flags:
            for i in range(n):
                if flags & (1<<i):
                    Flags.append(i)
            else:
                pass # flags == 0
    elif type(flags) in (tuple, list):
        Flags = flags
    else:
        print('type flags: %s'% type(flags))
        Flags = flags
    return tuple(Flags)

#GUI_CARETBLINKING = 1
#GUI_INMOVESIZE = 2
#GUI_INMENUMODE = 4
#GUI_SYSTEMMENUMODE = 8
#GUI_POPUPMENUMODE = 16
    

if __name__ == '__main__':
    # Quick test.
    #moveCursorInCurrentWindow(200, 200)
    # should return False:
    time.sleep(1)
    print('hasMenuOpen: %s'% hasMenuOpen())
    print('hasPopupOpen: %s'% hasPopupOpen())
    print('flags: %s'% repr(getWindowFlags()))